import os, subprocess, sys
from datetime import datetime
import program_utils
import file_utils

def run_main( pargs ):
    # time the duration of module container run
    run_start = datetime.now()
    print('Container running...')
    print('Program arguments: {}'.format(str(pargs)))
    # initialize program run
    # program_params = program_utils.init_program()

    # download any cloud files
    # pargs_updated = file_utils.download_files( dict(program_arguments=pargs) )

    # run main program
    # program_utils.run_program( pargs_updated )

    # create run log that includes program run duration
    run_end = datetime.now()
    # program_utils.log_run( dict(program_arguments=pargs,
    #                            run_duration=str(run_end - run_start)))

    # upload output data files
    # program_utils.upload_output()
    print('Container finished in {}'.format(str(run_end - run_start)))
    return


if __name__ == '__main__':
    run_main(sys.argv[1:]) if len(sys.argv) > 1 else run_main([])



import os, sys, subprocess, json, uuid

SCRIPT_DIR = str(os.path.dirname(os.path.realpath(__file__)))
HOME_PATH = os.path.expanduser('~')
INIT_PATH = os.path.join(HOME_PATH, '.bioshedinit/')
VALID_REGIONS = ['us-west-2', 'us-west-1', 'us-east-1', 'us-east-2', 'af-south-1', 'ap-east-1', \
                 'ap-southeast-3', 'ap-south-1', 'ap-northeast-3', 'ap-northeast-2', 'ap-southeast-1', 'ap-southeast-2', 'ap-northeast-1', \
                 'ca-central-1', 'eu-central-1', 'eu-west-1', 'eu-west-2', 'eu-south-1', 'eu-west-3', 'eu-north-1', \
                 'me-south-1', 'me-central-1', 'se-east-1']
VALID_REGION_NUMBERS = {'1': 'us-west-1', '2': 'us-west-2', '3': 'us-east-1', '4': 'us-east-2', '5': 'ap-south-1', \
                        '6': 'ap-northeast-1', '7': 'ap-northeast-2', '8': 'ap-southeast-1', '9': 'ca-central-1', '10': 'eu-west-1', '11': 'eu-west-2', \
                        '12': 'eu-west-3', '13': 'eu-central-1'}

ECR_PUBLIC_REGISTRY = "public.ecr.aws/w7q0j5w1"
BIOSHED_SERVERLESS_API = "https://hu9ug76w32.execute-api.us-west-2.amazonaws.com/prod"

sys.path.append(os.path.join(SCRIPT_DIR, 'bioshed_utils/'))
import quick_utils

def userExists( user ):
    """ Check if username exists in the user database.
    """
    mybody = {"user": user}
    user_exists = quick_utils.post_request(dict(url=BIOSHED_SERVERLESS_API+'/searchusers', body=mybody))
    if "user_found" in user_exists and str(user_exists["user_found"])[0].upper() == "T":
        return True
    else:
        return False

def bioshed_login():
    """ Log into BioShed.
    """
    myuser = input('Enter your BioShed username (email address): ') or ""
    user_exists = False
    if myuser != "":
        user_exists = userExists(myuser)
    if not user_exists:
        print("ERROR: User {} does not exist.".format(str(myuser)))
    return {"login": False, "user": myuser} if not user_exists else {"login": True, "user": myuser}

def bioshed_init( args ):
    """
    system: ubuntu, macosx,...
    cloud: aws, gcp,...
    initpath: path to all init and setup files
    configfile: config file for important config constants
    providerfile: TF provider file
    mainfile: TF main file
    ---
    provider_file

    """
    system_type = args['system']
    init_path = args['initpath']
    which_os = ''

    # first create init directory if doesn't exist
    if not os.path.exists(init_path):
        os.mkdir(init_path)

    if 'ubuntu' in system_type.lower():
        bioshed_init_ubuntu()
        which_os = 'ubuntu'
    elif 'debian' in system_type.lower():
        bioshed_init_ubuntu()
        which_os = 'debian'
    elif 'redhat' in system_type.lower():
        bioshed_init_redhat()
        which_os = 'redhat'
    elif 'centos' in system_type.lower():
        bioshed_init_redhat()
        which_os = 'centos'
    elif 'macosx' in system_type.lower():
        bioshed_init_macosx()
        which_os = 'macosx'
    elif 'amazon' in system_type.lower():
        bioshed_init_amazonlinux()
        which_os = 'amazonlinux'
    else:
        print('ERROR: Unsupported system OS - exiting.')

    return which_os

def bioshed_setup( args ):
    """
    cloud: aws, gcp,...
    initpath: path to all init and setup files
    configfile: config file for important config constants
    providerfile: TF provider file
    mainfile: TF main file
    ---
    provider_file

    [TODO] fix this: Found preexisting AWS CLI installation: /usr/local/aws-cli/v2/current. Please rerun install script with --update flag.
    """
    cloud_provider = args['cloud']
    init_path = args['initpath']
    config_file = args['configfile']
    provider_file = args['providerfile']
    main_file = args['mainfile']
    setup_again = 'Y'

    if cloud_provider.lower() in ['aws','amazon']:
        cloud_provider = 'aws'
        if cloud_setup( dict(cloud=cloud_provider, configfile=config_file) ):
            setup_again = PKEY_INPUT = input('Detected existing setup. Overwrite? (Y/N): ') or 'N'
        if setup_again.upper()[0] == 'Y':
            bioshed_init_aws()
            api_key_file = generate_api_key( dict(cloud=cloud_provider, configfile=config_file))
            provider_file = bioshed_setup_aws( dict(initpath=init_path, configfile=config_file, providerfile=provider_file, mainfile=main_file, keyfile=api_key_file))
            env_file = write_env_file( dict(cloud=cloud_provider, initpath=init_path))
            print('\nBioShed AWS integration setup successful! To setup the core AWS resources, now type:')
            print('\t$ bioshed deploy core')
            print('\n Or try one of the following to test your setup:')
            print('\t$ bioshed run fastqc --help')
            print('\t$ bioshed run fastqc --example')
    return provider_file

def bioshed_init_macosx():
    # first make sure brew is installed: https://brew.sh/
    # to install pip: subprocess.call('brew install brew-pip', shell=True)
    # install terraform
    if int(subprocess.call('brew --help', shell=True)) != 0:
        subprocess.call('/bin/bash -c "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/HEAD/install.sh)"', shell=True)
        # brew update ?
        # brew upgrade ?
    if int(subprocess.call('terraform --help', shell=True)) != 0:
        subprocess.call('brew tap hashicorp/tap', shell=True)
        subprocess.call('brew install hashicorp/tap/terraform', shell=True)
    if int(subprocess.call('docker --help', shell=True)) != 0:
        subproess.call('brew cask install docker', shell=True)
    return

def bioshed_init_amazonlinux():
    # install pip
    subprocess.call('sudo yum -y update', shell=True)
    subprocess.call('sudo yum -y install python3-pip zip unzip', shell=True)
    subprocess.call('sudo yum install -y yum-utils', shell=True)

    # install terraform: https://learn.hashicorp.com/tutorials/terraform/install-cli?in=terraform/aws-get-started
    if int(subprocess.call('terraform --help', shell=True)) != 0:
        subprocess.call('sudo yum-config-manager --add-repo https://rpm.releases.hashicorp.com/AmazonLinux/hashicorp.repo', shell=True)
        subprocess.call('sudo yum -y install terraform', shell=True)

    # install docker
    if int(subprocess.call('docker --help', shell=True)) != 0:
        subprocess.call('sudo yum install docker', shell=True)
    return

def bioshed_init_redhat():
    # install pip
    subprocess.call('sudo yum -y update', shell=True)
    subprocess.call('sudo yum -y install python3-pip zip unzip', shell=True)

    # install terraform: https://learn.hashicorp.com/tutorials/terraform/install-cli?in=terraform/aws-get-started
    subprocess.call('sudo yum install -y yum-utils', shell=True)
    subprocess.call('sudo yum-config-manager --add-repo https://rpm.releases.hashicorp.com/RHEL/hashicorp.repo', shell=True)
    subprocess.call('sudo yum -y install terraform', shell=True)

    # install docker
    if int(subprocess.call('docker --help', shell=True)) != 0:
        subprocess.call('sudo yum install docker', shell=True)
    return

def bioshed_init_ubuntu():
    # install pip
    subprocess.call('sudo apt-get -y update', shell=True)
    subprocess.call('sudo apt-get -y install python3-pip zip unzip', shell=True)

    # install terraform: https://learn.hashicorp.com/tutorials/terraform/install-cli?in=terraform/aws-get-started
    subprocess.call('sudo apt-get install -y gnupg software-properties-common', shell=True)
    subprocess.call('wget -O- https://apt.releases.hashicorp.com/gpg | gpg --dearmor | sudo tee /usr/share/keyrings/hashicorp-archive-keyring.gpg', shell=True)
    subprocess.call('gpg --no-default-keyring --keyring /usr/share/keyrings/hashicorp-archive-keyring.gpg --fingerprint', shell=True)
    subprocess.call('echo "deb [signed-by=/usr/share/keyrings/hashicorp-archive-keyring.gpg] https://apt.releases.hashicorp.com $(lsb_release -cs) main" | sudo tee /etc/apt/sources.list.d/hashicorp.list', shell=True)
    subprocess.call('sudo apt update', shell=True)
    subprocess.call('sudo apt-get install terraform', shell=True)

    # boto3 for python-based infra control
    # subprocess.call('pip install boto3', shell=True)

    # install docker
    if int(subprocess.call('docker --help', shell=True)) != 0:
        subprocess.call('sudo apt-get install -y docker.io', shell=True)
        # subprocess.call('sudo snap install docker', shell=True)

    return

def bioshed_init_aws():
    # install aws cli and configure
    if int(subprocess.call('aws --help', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)) > 3:
        subprocess.call('pip install awscli', shell=True)
        subprocess.call('curl "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "awscliv2.zip"', shell=True)
        subprocess.call('unzip awscliv2.zip', shell=True)
        subprocess.call('sudo ./aws/install', shell=True)
    subprocess.call('aws configure', shell=True)

def bioshed_setup_aws( args ):
    """
    initpath: init path
    configfile: config file for important config constants
    providerfile: TF provider file
    mainfile: TF main file
    keyfile: API key file
    ---
    """
    # setup terraform provider (AWS): https://learn.hashicorp.com/tutorials/terraform/aws-build
    cwd = os.getcwd()
    INIT_PATH = args['initpath']
    if not os.path.exists(INIT_PATH):
        os.mkdir(INIT_PATH)
    apikeyfile = args['keyfile'] if 'keyfile' in args else ''
    PROVIDER_FILE = args['providerfile'] # os.path.join(INIT_PATH, 'hs_providers.tf')
    MAIN_FILE = args['mainfile'] # os.path.join(INIT_PATH, 'main.tf')
    AWS_CONFIG_FILE = args['configfile']
    KEYS_FILE = os.path.join(INIT_PATH, 'hsconfig.tf')
    AWS_REGION = input('Which region do you want to setup your bioshed infrastructure?\n1: us-west-1 (california)\n2: us-west-2 (oregon) - COMMONLY USED\n3: us-east-1 (virginia) - COMMONLY USED\n4: us-east-2 (ohio)\n5: ap-south-1 (india) - COMMONLY USED\n6: ap-northeast-1 (tokyo)\n7: ap-northeast-2 (seoul)\n8: ap-southeast-1 (singapore)\n9: ca-central-1 (canada)\n10: eu-west-1 (ireland) - COMMONLY USED\n11: eu-west-2 (london)\n12: eu-west-3 (paris)\n13: eu-central-1 (germany)\n\nType a number or a region name: ')
    if AWS_REGION not in VALID_REGIONS and AWS_REGION not in VALID_REGION_NUMBERS:
        AWS_REGION = input('Please type a valid number or AWS region (default: us-west-2): ') or "us-west-2"
    if AWS_REGION in VALID_REGION_NUMBERS:
        AWS_REGION = VALID_REGION_NUMBERS[AWS_REGION]
    AWS_CONSTANTS_JSON = {}
    AWS_CONSTANTS_JSON["ecr_registry"] = ECR_PUBLIC_REGISTRY
    AWS_CONSTANTS_JSON["aws_region"] = AWS_REGION

    if apikeyfile == '':
        PKEY_INPUT = input('Provide a valid public key for accessing AWS resources (in a separate window, type "bioshed keygen aws" or type "ssh-keygen" and paste the public key here). Press ENTER to skip if not using AWS resources: ')
    else:
        PKEY_INPUT = get_public_key( dict(configfile=AWS_CONFIG_FILE))

    with open(PROVIDER_FILE,'w') as f:
        f.write('terraform {\n')
        f.write('  required_providers {\n')
        f.write('    aws = {\n')
        f.write('      source  = "hashicorp/aws"\n')
        f.write('      version = "~> 3.0"\n')
        f.write('    }\n')
        f.write('  }\n')
        f.write('}\n')
        f.write('\n')
        f.write('provider "aws" {\n')
        f.write('  region = "{}"\n'.format(AWS_REGION))
        f.write('  shared_credentials_file = "$HOME/.aws/credentials"\n')
        f.write('}\n')
    with open(MAIN_FILE,'w') as f:
        f.write('\n')
    if PKEY_INPUT not in ['',' ',[]]:
        with open(KEYS_FILE,'w') as f:
            f.write(
            """
            resource "aws_key_pair" "deployer" {
              key_name   = "bioshed-managed-key-for-ec2-instances"
            """
            )
            f.write('  public_key = "{}"\n'.format(PKEY_INPUT))
            f.write('}\n\n')
    AWS_CONSTANTS_JSON['setup'] = 'True'
    quick_utils.add_to_json( AWS_CONFIG_FILE, AWS_CONSTANTS_JSON)
    # with open(AWS_CONFIG_FILE,'w') as f:
    #     json.dump(AWS_CONSTANTS_JSON, f)
    os.chdir(INIT_PATH)
    # initialize terraform
    subprocess.call('terraform init', shell=True)
    os.chdir(cwd)
    return PROVIDER_FILE

def bioshed_teardown( args ):
    """ Destroys cloud infrastructure setup by bioshed.
    Be very careful before doing this.

    initpath: init path
    """
    cwd = os.getcwd()
    INIT_PATH = args['initpath']
    os.chdir(INIT_PATH)
    subprocess.call('terraform plan -destroy', shell=True)
    subprocess.call('terraform apply -destroy', shell=True)
    os.chdir(cwd)
    return

def write_env_file( args ):
    """ Writes environment file for use with cloud provider containers.

    cloud: cloud_provider
    initpath: path of config and init files for bioshed
    group: key group if multiple credentials are in credentials file. Default: 'default'
    ---
    envfile: path and name of env file
    """
    cloud = args['cloud'] if 'cloud' in args else 'aws'
    initpath = args['initpath'] if 'initpath' in args else ''
    HOME_PATH = os.path.expanduser('~')
    group = args['group'] if 'group' in args else 'default'
    envfile = os.path.join(initpath,'.env_{}'.format(cloud))
    credfile = ''
    access_key = ''
    secret_key = ''
    current_group = ''
    if initpath != '' and os.path.exists(initpath):
        if cloud in ['aws','amazon']:
            credfile = os.path.join(HOME_PATH,'.aws/credentials')
            if os.path.exists(credfile):
                with open(credfile,'r') as f:
                    r = 'start'
                    while r != '':
                        r = f.readline()
                        if 'aws_access_key_id' in r and current_group == group:
                            access_key = str(r.strip().split('=')[-1]).strip()
                        elif 'aws_secret_access_key' in r and current_group == group:
                            secret_key = str(r.strip().split('=')[-1]).strip()
                        elif '[' in r and ']' in r:
                            if group in r:
                                current_group = group
                            else:
                                current_group = ''
            with open(envfile,'w') as fout:
                fout.write('AWS_ACCESS_KEY_ID={}\n'.format(access_key))
                fout.write('AWS_SECRET_ACCESS_KEY={}\n'.format(secret_key))
    return envfile

def get_env_file( args ):
    """ Gets environment file for cloud provider.

    cloud: cloud provider
    initpath: init path with all config and env files
    ---
    envfile: env file for that cloud provider
    """
    cloud = args['cloud'] if 'cloud' in args else 'aws'
    initpath = args['initpath'] if 'initpath' in args else ''
    if cloud=='aws' and initpath != '':
        return os.path.join(initpath,'.env_{}'.format(cloud))
    else:
        return ''

def cloud_setup( args ):
    """ Checks if cloud provider already setup in Bioshed. (bioshed setup [cloud])
    Assumes that config file has a key 'setup = True' or 'setup' = 'Yes' if setup properly.

    cloud: which cloud provider
    configfile: config file for cloud provider
    ---
    boolean: True/False is cloud setup already?

    """
    isSetup = False
    cloud = args['cloud'] if 'cloud'in args else 'aws'
    configfile = args['configfile'] if 'configfile' in args else ''
    if os.path.exists(configfile):
        config_json = quick_utils.loadJSON(configfile)
        is_cloud_setup = config_json['setup'] if 'setup' in config_json else 'False'
        if is_cloud_setup.upper()[0] in ['T','Y']:
            isSetup = True
    return isSetup

def cloud_configured( args ):
    """ Checks if a cloud provider is configured properly

    cloud: which cloud provider to check. If empty, then check all.
    ---
    boolean: True/False if cloud is configured or not
    """
    cloud = args['cloud'] if 'cloud' in args else 'all'
    pnum = 127
    if cloud in ['aws','amazon','all']:
        pnum = min(int(subprocess.call('aws s3 ls', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)), pnum)
    if cloud in ['gcp','all']:
        pnum = min(int(subprocess.call('gcp --help', shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)), pnum)

    if pnum > 3:
        return False
    else:
        return True

def cloud_core_setup( args ):
    """ Checks if cloud core is setup (i.e., bioshed deploy core) has been run).

    configfile: config file for core setup.
    ---
    boolean: True/False if cloud is configured or not
    """
    isCoreSetup = False
    configfile = args['configfile'] if 'configfile' in args else ''
    if configfile!='' and os.path.exists(configfile):
        config_json = quick_utils.loadJSON(configfile)
        is_core_setup = config_json['core_setup'] if 'core_setup' in config_json else 'N'
    if is_core_setup.upper()[0] in ['Y','T']:
        isCoreSetup = True
    return isCoreSetup

def generate_api_key( args ):
    """ Generates an API key for a cloud provider.

    cloud: which cloud provider to generate key for.
    configfile: cloud provider config file
    ---
    keyfile: name of generated public key file

    [NOTE] https://stackoverflow.com/questions/43235179/how-to-execute-ssh-keygen-without-prompt

    """
    cloud = args['cloud'] if 'cloud' in args else 'aws'
    configfile = args['configfile'] if 'configfile' in args else ''
    CONFIG_JSON = {}
    unique_key_id = str(uuid.uuid4())[0:4]
    keyfile = os.path.join(INIT_PATH, 'bioshed_{}_{}'.format(cloud, unique_key_id))
    while os.path.exists(keyfile):
        # in the rare event that key with unique id already exists, generate a new one
        unique_key_id = str(uuid.uuid4())[0:4]
        keyfile = os.path.join(INIT_PATH, 'bioshed_{}_{}'.format(cloud, unique_key_id))
    if configfile != '':
        # generate a new public/private key pair for cloud API.
        rcode = subprocess.call("ssh-keygen -t rsa -N '' -f {}".format(keyfile), shell=True)
        CONFIG_JSON['apikeyfile'] = keyfile
        if int(rcode) == 0:
            print('Public/private key generated at {}'.format(keyfile))
        # reference key file name within config file
        quick_utils.add_to_json( configfile, CONFIG_JSON)
    else:
        print('ERROR: No key generated. Must specify a cloud config file.')
    return keyfile

def get_public_key( args ):
    """ Gets public key for a given key file specified within config file.

    configfile: config file where key file name is stored
    ---
    pubkey: public key
    """
    configfile = args['configfile']
    keyfile = ''
    pubkey = ''

    if os.path.exists(configfile):
        configjson = quick_utils.loadJSON(configfile)
        keyfile = configjson['apikeyfile']
    if keyfile != '' and os.path.exists(keyfile+'.pub'):
        with open(keyfile+'.pub','r') as f:
            # takes last line as key
            for r in f:
                pubkey = r.strip()
    return pubkey

def bioshed_run_help():
    """ Help menu for bioshed run.
    """
    print("""

    ------------------------------------------------------------
    BioShed RUN
    ------------------------------------------------------------
    Welcome to BioShed Run! With this tool, you can:
        - Run bioinformatics applications as you would normally, but without the hassle of installation or setup.
        - Perform bioinformatics on your own computer or in the cloud.
        - Seamlessly work with data files on local or remote cloud file systems.

    Usage:
        bioshed run <optional:CONFIG-ARGS> <APP> <PROGRAM-ARGS> out::OUTPUT_DIR(optional)

    <optional CONFIG-ARGS>
        --local                 Explicitly run on local system. Requires docker installed and running
        --aws-env-file <.ENV>   AWS environment file with the following format:
                                  AWS_ACCESS_KEY_ID=[ID]
                                  AWS_SECRET_ACCESS_KEY=[KEY]
                                  AWS_DEFAULT_REGION=[region]
        --inputdir <DIR>        Full path of local input directory. Requires --local. By default, the current directory is used.


    <APP>
        Type "bioshed run --list" to see a list of available applications. Examples are:
            fastqc              FastQC Quality Control for sequencing data
            STAR                RNA-STAR aligner
            bwa                 Burrows-Wheeler aligner
            gatk                GATK sequencing workflow toolkit

        You can also run BioContainers on BioShed. For help with this, just type:
            $ bioshed run biocontainers --help

    <PROGRAM-ARGS>
        Run any bioinformatics application as you would normally.
        One special case is:
            /input/<FILE>       When running locally, prefix any local files with /input/ - ex: /input/my.fastq


    out::OUTPUT_DIR (optional)
        out::<OUTPUT_DIR>       Optional output directory for all data + log files. Specify full path (local or remote).
                                Without this option, files are output back to the input path.
        out::local              Special case - use the current local directory as the output directory.


    The following files are output:
        bioshed.run.out         All text printed to the console (STDOUT)
        program.cat.<TIME>.log  Log file of program output
        run.<TIME>.log          Run log JSON
        out::<OUTPUT_FOLDER>    Output folder for all log and data files. You can explicitly specify local by "out::local"
                                If no output folder is specified, then the input folder is used as the output folder.

    ------------------------------------------------------------
    EXAMPLES
    ------------------------------------------------------------
        Using cloud storage:
            $ bioshed run fastqc s3://folder1/seq.fastq.gz
            $ bioshed run bwa mem s3://genomes/bwa_index s3://folder1/seq.fastq.gz out://s3://alignments/
            $ bioshed run STAR --genomeDir s3://genomes/hg38_STAR_index/ --readFilesIn s3://fastqs/my.fastq.gz out::s3://alignments/

        Using local storage:
            $ bioshed run zcat /input/my.fastq.gz out::local
            $ bioshed run fastqc file.fastq.gz out::local
            $ bioshed run --local --inputdir /home/ cat /input/README.txt

        Using local and cloud (hybrid) storage:
            $ bioshed run bedtools merge s3://folder1/test.bed out::/data/
            $ bioshed run fastqc s3://bioshed-examples/fastq/rnaseq_mouse_test_tiny1_R1.fastq.gz out::local

    ------------------------------------------------------------
    HELP
    ------------------------------------------------------------
        For help on how to run an application, type one of the following:
            $ bioshed run <APP> --help
            $ bioshed run <APP> --example

        Examples:
            $ bioshed run fastqc --help     Shows help menu for FASTQC - output to local file "bioshed.run.out"
            $ bioshed run fastqc --example  Shows an example of running FASTQC - output to local file "bioshed.run.out"

    """)

def biocontainers_help():
    """ Help menu for bioshed run biocontainers.
    """
    print("""
    BioShed allows you to run BioContainers without knowing details about Docker.

    To run a basic biocontainer, type (for example):
    $ bioshed run biocontainers blast:2.2.31 makeblastdb --help

    The basic format is:
    $ bioshed run biocontainers <IMAGE>:<TAG> <COMMAND>

    You must know the image name and tag of the biocontainer you want to run.
    For more information, refer to the Biocontainers documentation at:
    https://biocontainers.pro/

    """)



import sys, os, json
import pandas as pd
SCRIPT_DIR = str(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(os.path.join(SCRIPT_DIR))
import atlas_utils
sys.path.append('bioshed_utils/')
import quick_utils
import aws_s3_utils

DEFAULT_SEARCH_FILE = os.path.join(os.getcwd(),"search_encode.txt")

def search_encode( args ):
    """ Entrypoint for an ENCODE search.
    $ bioshed search encode <searchterms>
    Examples
    $ bioshed search encode breast cancer rna-seq
    $ bioshed search encode --tissue heart --assay chip-seq

    searchterms: search terms input by user
    ---
    results: data frame of results

    >>> search_encode( dict(searchterms='breast cancer rna-seq'))
    ''
    >>> search_encode( dict(searchterms='--tissue heart --assay chip-seq'))
    ''
    >>> search_encode( dict(searchterms='--assay single cell --tissue heart'))
    ''

    Prints number of experiment datasets found and where results are output to (search_encode.txt).
    Outputs search results to search_encode.txt (tab-delimited text)
    This search results file is fed into download relevant files with the following command:

    $ bioshed download encode

    You can further filter search_encode.txt before download by adding search terms:

    $ bioshed download encode --filetype fastq

    By default, this will download to the current folder. You can specify a relative path to download or a remote cloud bucket to download to with the --output parameter.

    $ bioshed download encode --output <local_outdir>
    $ bioshed download encode --output s3://my/s3/folder

    NOTE: You MUST have a bioshed_encode.txt file before you run bioshed download.

    You can also specify a bioshed-formatted ENCODE results file to download files:

    $ bioshed download encode newsearch_results.txt

    For help with anything, type:
    $ bioshed search encode --help
    $ bioshed download encode --help
    """
    URL_BASE = 'https://encodeproject.org/search/'
    url_search_string = ''
    search_results = {}
    # dictionary of search terms: {"general": "...", "tissue": "...", "celltype": "..."...}
    search_dict = atlas_utils.parse_search_terms( args['searchterms'] ) if ('searchterms' in args and args['searchterms'] != '') else {}
    if search_dict == {} or 'help' in search_dict:
        print_encode_help()
    else:
        # start with search url base and build it according to search terms
        for category, terms in search_dict.items():
            url_search_string = combine_search_strings(url_search_string, convert_to_search_string( dict(terms=terms, category=category)))
        if url_search_string != '' and '&searchTerm=&' not in url_search_string:
            search_results = encode_search_url( dict(url='/search/{}'.format(url_search_string), searchtype='full', returntype='full'))
    return search_results

def encode_search_url( args ):
    """ Searches ENCODE by a URL suffix.
    https://www.encodeproject.org/<URL_SUFFIX>

    url: url suffix to search
    searchtype: 'full' (full search), 'experiment' (search within an experiment), 'file' (search within a file)
    returntype: 'full' (default), 'raw', 'file', 'experiment', 'celltype', 'assay', 'platform',...
    ---
    results:

    - By default, this returns the full table of info, and also writes results to a default file.
    - We can also return the raw JSON result with 'raw' as the return type.
    - If the return type is a specific column of the full table, then a
      "joined" table is returned where the key is the requested return type.
        ex: returntype='file'
        FILE    EXPERIMENT  CELLTYPE    ASSAY   ACCESSION
        ... (one row per file)
    [NOTE] search is limited to 50000 results

    >>> encode_search_url(dict(url="experiments/ENCSR000BDC/", searchtype="experiment", returntype="raw"))
    ''
    >>> encode_search_url(dict(url="experiments/ENCSR000BDC/", searchtype="experiment", returntype="file"))
    ''
    """
    returntype = args['returntype'] if 'returntype' in args else 'full'
    searchtype = args['searchtype'] if 'searchtype' in args else 'full'
    if 'url' not in args:
        print('ERROR: You need to specify a URL.')
        return {}
    elif str(args['url']).lstrip('/').startswith('search'):
        search_url = 'https://www.encodeproject.org/{}&limit=50000'.format(str(args['url']).lstrip('/'))
    else:
        search_url = 'https://www.encodeproject.org/{}'.format(str(args['url']).lstrip('/'))

    print('GET request: {}'.format(search_url))
    results_raw = quick_utils.get_request( dict(url=search_url, type='application/json'))
    if returntype.lower() == 'raw':
        return results_raw
    elif searchtype.lower() == 'full':
        return get_full_info_from_encode_json( dict(results=results_raw, sortby=returntype, search_string=search_url))
    elif returntype.lower() == 'experiment':
        return get_experiments_from_encode_json( dict(results=results_raw))
    elif returntype.lower() == 'file':
        return get_files_from_encode_json( dict(results=results_raw, searchtype=searchtype))
    else:
        return results_raw

def get_full_info_from_encode_json( args ):
    """ Get experiments with full info from an ENCODE search JSON.
    results: results_raw
    sortby: which column to sort by - full/experiment (default), file, assay, etc...
    returntype: JSON or dataframe (default)
    search_string: original search string
    ---
    fullinfo: JSON or dataframe
    DEFAULT_SEARCH_FILE (outfile): table

    https://www.encodeproject.org/search/?type=Experiment&searchTerm=breast+cancer
    FOR EACH EXPERIMENT:
        "assay_term_name": "ChIP-seq",
        "assay_title": "Control ChIP-seq",
        "biosample_ontology": {
            "term_name": "MCF-7"
        },
        "biosample_summary": "Homo sapiens MCF-7",
        "dbxrefs": [
            "GEO:GSM1010854", <--
            "UCSC-ENCODE-hg19:wgEncodeEH003428"
        ],

    """
    results = args['results']
    sortby = args['sortby'] if 'sortby' in args else 'experiment'
    returntype = args['returntype'] if 'returntype' in args else 'pandas'
    search_string = args['search_string'] if 'search_string' in args else ''

    if sortby in ['full', 'experiment']:
        tbl = {"experiment": [], "assay": [], "celltype": [], "species": [], "accession": [], "file": []}
        for fullexpt in results["@graph"]:
            if "@id" in fullexpt:
                tbl['experiment'].append(str(fullexpt["@id"]))
                tbl['assay'].append(str(fullexpt["assay_term_name"]) if "assay_term_name" in fullexpt else '')
                tbl['celltype'].append(str(fullexpt["biosample_ontology"]["term_name"]) if "biosample_ontology" in fullexpt and "term_name" in fullexpt["biosample_ontology"] else '')
                tbl['species'].append(' '.join(str(fullexpt["biosample_summary"]).split(' ')[0:2]) if "biosample_summary" in fullexpt else '')
                tbl['accession'].append(list(fullexpt["dbxrefs"]) if "dbxrefs" in fullexpt else [])
                tbl['file'].append(list(map(lambda f: f["@id"], fullexpt["files"])) if "files" in fullexpt else [])
        print('Number of experiment datasets found: {}'.format(str(len(tbl['experiment']))))
        print('Number of assays found: {}'.format(str(len(list(set(tbl['assay']))))))
        print('Number of cell types found: {}'.format(str(len(list(set(tbl['celltype']))))))
        print('Number of total files found: {}'.format(str(sum([len(e) for e in tbl['file']]))))
    elif sortby == 'assay':
        # create joined table for gathering info
        jtbl = {}  # key is assay
        for fullexpt in results["@graph"]:
            if "@id" in fullexpt and "assay_term_name" in fullexpt:
                assay = fullexpt["assay_term_name"]
                if assay not in jtbl:
                    jtbl[assay] = {"experiment": [], "celltype": [], "species": [], "accession": [], "file": []}
                jtbl[assay]["experiment"].append(str(fullexpt["@id"]))
                if "biosample_ontology" in fullexpt and "term_name" in fullexpt["biosample_ontology"]:
                    jtbl[assay]['celltype'].append(str(fullexpt["biosample_ontology"]["term_name"]))
                if "biosample_summary" in fullexpt:
                    jtbl[assay]['species'].append(' '.join(str(fullexpt["biosample_summary"]).split(' ')[0:2]))
                if "dbxrefs" in fullexpt:
                    jtbl[assay]['accession'].append(list(fullexpt["dbxrefs"]))
                if "files" in fullexpt:
                    jtbl[assay]['file'].append(list(map(lambda f: f["@id"], fullexpt["files"])))
        # create flat table for output
        tbl = {"assay": [], "experiment": [], "celltype": [], "species": [], "accession": [], "file": []} # full table
        for k, v in jtbl.items():
            tbl["assay"].append(k)
            tbl["experiment"].append(v["experiment"] if "experiment" in v else '')
            tbl["celltype"].append(v["celltype"] if "celltype" in v else '')
            tbl["species"].append(v["species"] if "species" in v else '')
            tbl["accession"].append(v["accession"] if "accession" in v else [])
            tbl["file"].append(v["file"] if "file" in v else [])

    if returntype in ['pandas', 'dataframe']:
        with open(DEFAULT_SEARCH_FILE,'w') as fout:
            fout.write('# bioshed search encode {}\n'.format(search_string))
        tbl_df = pd.DataFrame(tbl)
        tbl_df.index.name = 'index'
        tbl_df.to_csv(DEFAULT_SEARCH_FILE, sep='\t', mode='a')
        print('Search results written to {}.'.format(DEFAULT_SEARCH_FILE.split('/')[-1]))
        print('Type "bioshed download encode" to download data files or "bioshed download encode --list" for file info before downloading.')
        return tbl_df
    else:
        return tbl

def get_experiments_from_encode_json( args ):
    """ Get experiment IDs from an ENCODE search JSON.
    results: results JSON from a raw search.
    ---
    expts: list of experiments in URL format (/experiments/ENCSR000ACHE/,...)

    https://www.encodeproject.org/experiments/ENCSR000AHE/
    """
    results = args['results']
    expts = []
    for fullexpt in results["@graph"]:
        expts.append(str(fullexpt["@id"]))
    return expts

def get_files_from_encode_json( args ):
    """ Get files from an ENCODE results JSON
    results: results JSON from a raw ENCODE search
    searchtype: the type of raw ENCODE search passed in
    cloud: which remote cloud (aws/s3 or google/gcp or microsoft/azure)
    ---
    relevant_files: list of S3 file URIs

    [TODO] for 'file' searchtype, can get HTTPS object link via 'cloud_metadata' key.
    [TODO] deal with JSON entries that don't have s3_uri (e.g., SRA files) - GET request: https://www.encodeproject.org/experiments/ENCSR860HAA/
    """
    results = args['results']
    searchtype = args['searchtype'] if 'searchtype' in args else ''
    cloud = args['cloud'] if 'cloud' in args else 's3'

    relevant_files = []
    if searchtype in ['experiment']:
        # original search was an experiment
        if "files" in results:
            for f in results["files"]:
                if cloud in ['s3','aws','amazon']:                    
                    if "s3_uri" in f and quick_utils.cloud_initialized(dict(cloud='aws')):
                        relevant_files.append(f["s3_uri"])
                    elif "cloud_metadata" in f and "url" in f["cloud_metadata"]:
                        relevant_files.append(f["cloud_metadata"]["url"])
    elif searchtype in ['file']:
        if "s3_uri" in results and quick_utils.cloud_initialized(dict(cloud='aws')):
            relevant_files.append(results["s3_uri"])
        elif "cloud_metadata" in f and "url" in f["cloud_metadata"]:
            revelant_files.append(results["cloud_metadata"]["url"])        
    return relevant_files

def download_encode( args ):
    """ Entrypoint for an ENCODE download.
    Assumes that search_encode() has already been run, so that a search_encode.txt file exists.
    User can refine search.
    downloadstr: download string passed in, which can include the following:

    --help (help menu)
    --list : list files, but do not download (like a dryrun)
    --update : update only - do not overwrite existing files
    --input <file name>
    --output <output directory>
    --filetype <refine by file type(s) - space delimited>
    --assay <refine by assay>
    --species <refine by species>
    --experiment <refine by experiment ID>
    --celltype <refine by cell type>
    --index <refine by index>

    [TODO] Clean up documentation and write tests

    [NOTE] Current format for search_encode.txt is:
    index	experiment	assay	celltype	species	accession	file
    0	/experiments/ENCSR718YPN/	single-nucleus ATAC-seq	heart left ventricle	Homo sapiens	[]	['/files/ENCFF804ONU/', '/files/ENCFF393IGF/',...]

    [NOTE] Use str.contains:  df2 = df.loc[df['celltype'].str.contains('heart', case=False)]
    [DONE] s3 file transfer function in aws_s3_utils
    [TODO] create https to s3 transfer, https to gcp transfer, gcp to s3 transfer etc
    """
    INFO_COLUMNS = ['experiment', 'assay', 'celltype', 'species']
    dd = atlas_utils.parse_search_terms( args['downloadstr'] if 'downloadstr' in args else '')
    infile = dd['input'] if 'input' in dd else 'search_encode.txt'
    outdir = dd['output'] if 'output' in dd else str(os.getcwd())
    filetype = dd['filetype'] if 'filetype' in dd else ''
    assay = dd['filetype'] if 'filetype' in dd else ''
    species = dd['species'] if 'species' in dd else ''
    experiment = dd['experiment'] if 'experiment' in dd else ''
    celltype = dd['celltype'] if 'celltype' in dd else ''
    index = dd['index'] if 'index' in dd else ''
    listonly = 'True' if 'list' in dd else ''
    updateonly = 'True' if 'update' in dd else ''
    outfiles = []
    outfiles_info = {}
    downloaded_files = []
    eids = []
    annotation_info = []
    ANNOTATION_INFO_FILE = os.path.join(os.getcwd(),'annotation_encode.txt')
    original_search_command = ''

    if 'help' in dd:
        print_encode_help()
    elif os.path.exists(infile):
        # get comment line (first line), if there
        with open(infile,'r') as f:
            r = f.readline()
            if r[0] == '#':
                original_search_command = r.strip()
        # get table
        df = pd.read_csv(infile, sep='\t', comment='#')
        if assay != '':
            df = df.loc[df['assay'].str.contains(assay, case=False)]
        if species != '':
            df = df.loc[df['species'].str.contains(species, case=False)]
        if experiment != '':
            # if --experiment filter is specified
            eids = list(map(lambda e: '/experiments/{}/'.format(e), quick_utils.format_type(experiment, 'list')))
            # df = df.loc[df['experiment'].str.lower().contains(experiment, case=False)]
        if celltype != '':
            df = df.loc[df['celltype'].str.lower().contains(celltype, case=False)]
        if index != '':
            df = df.loc[df['index'].str.lower().contains(index, case=False)]

        experiment_urls = list(df['experiment'])
        for e_url in experiment_urls:
            if len(eids)==0 or (len(eids)>0 and e_url in eids):
                # get paths of all files for each experiment
                outfiles_new = encode_search_url( dict(url=e_url, searchtype='experiment', returntype='file'))
                outfiles += outfiles_new
                for outfile in outfiles_new:
                    outfiles_info[outfile] = '; '.join(list(map(str, df.loc[df['experiment']==e_url][INFO_COLUMNS].values.flatten().tolist())))

        if filetype!='':
            # if --filetype filter is specified
            ftypes = quick_utils.format_type(filetype, 'list')
            for ftype in ftypes:
                outfiles = list(filter(lambda f: ftype in f, outfiles))
                removed_files = list(filter(lambda f: ftype not in f, outfiles))
                for removed_file in removed_files:
                    removed_info = outfiles_info.pop(removed_file, 'not found')

        # gather annotation info
        annotation_info.append('FILE\tINFO (EXPT; ASSAY; CELLTYPE; SPECIES)')
        for outfile, outfile_info in outfiles_info.items():
            annotation_info.append('{}\t{}'.format(quick_utils.get_file_only(outfile), str(outfile_info)))

        if listonly == 'True':
            # list files, but do not download
            for annot in annotation_info:
                print(annot)
        else:
            # download files
            if outdir.startswith('s3') and len(outfiles) > 0 and outfiles[0].startswith('s3'):
                # s3-to-s3 file transfer
                downloaded_files = aws_s3_utils.transfer_file_s3( dict(path=outfiles, outpath=outdir, overwrite='False' if updateonly=='True' else 'True'))
            elif not outdir.startswith('s3') and len(outfiles) > 0 and outfiles[0].startswith('s3'):
                # s3 download to local
                downloaded_files = aws_s3_utils.download_file_s3( dict(path=outfiles, localdir=outdir, overwrite='False' if updateonly=='True' else 'True'))
            elif not outdir.startswith('s3') and len(outfiles) > 0 and outfiles[0].startswith('http'):
                # http download to local
                for outfile in outfiles:
                    print('Downloading {}'.format(outfile))                
                    # get http(s) file    
                    _response = quick_utils.get_request( dict(url=outfile))
                    # write to local file
                    open(os.path.join(outdir, quick_utils.get_file_only(outfile)), 'wb').write(_response.content)
                    # save list of downloaded files
                    downloaded_files.append(os.path.join(outdir, quick_utils.get_file_only(outfile)))
            
            # write out annotation info
            with open(ANNOTATION_INFO_FILE,'w') as fout:
                if original_search_command != '':
                    fout.write(original_search_command+'\n')
                for row in annotation_info:
                    datafile = row.strip().split('\t')[0]
                    annots = str(row.strip().split('\t')[-1]).lstrip('INFO (').rstrip(')').split(';')
                    fout.write('{}\t{}\n'.format(datafile, '\t'.join(annots)))
    else:
        print('File {} does not exist. Please first run "bioshed search encode"'.format(infile))
    return downloaded_files


########################## HELPER FUNCTIONS ############################

def print_dataframe( df ):
    """ Print pandas dataframe to command line
    """
    pd.set_option('display.expand_frame_repr', False)
    pd.set_option('display.max_colwidth', 50)
    pd.set_option('display.precision', 2)
    print(df)
    return

def print_encode_help():
    """ Prints the command-line help menu when a user types a wrong input or
    when a user just types "bioshed search encode".
    """
    print('Welcome to ENCODE dataset search and download, powered by BioShed Atlas.\n')
    
    print('------------------------------------------------------------')
    print('BioShed ENCODE SEARCH')
    print('------------------------------------------------------------\n')

    print('Usage:')
    print('\tbioshed search encode <SEARCH>\n')

    print('Search Examples:\n')
    print('\tGeneral search:\t\t$ bioshed search encode breast cancer rna-seq')
    print('\tSearch by category:\t$ bioshed search encode --tissue heart --assay chip-seq\n')

    print('Search categories:')
    print('\t--tissue\tTissue where DNA/RNA/protein material was extracted.')
    print('\t--celltype\tCell type where DNA/RNA/protein material was extracted.')
    print('\t--assay\t\tAssay type performed.') 
    print('\t--assaytarget\tAssay target gene or protein.')
    print('\t--disease\tDiagnosed disease type.')
    print('\t--filetype\tType of data file.')
    print('\t--platform\tInstrument platform used.')
    print('\t--species\tHuman, mouse, etc.')
    print('\t--genome\thg38, hg19, mm10, etc.\n')

    print('To list search terms (example):\n')
    print('\t$ bioshed search encode --tissue')
    print('')
    print('BioShed will write SEARCH results to a file "search_encode.txt" in the current directory.\n')

    print('------------------------------------------------------------')
    print('BioShed ENCODE DOWNLOAD')
    print('------------------------------------------------------------\n')
    print('Usage: To download data files listed in "search_encode.txt", type\n')
    print('\t$ bioshed download encode\n')

    print('Further refine the files you want using any of the search categories. For example\n')
    print('\t$ bioshed download encode --filetype fastq')
    print('\t$ bioshed download encode --experiment ENCSR597YMV ENCSR562QQA ENCSR568FZM\n')
    
    print('You can list the files before downloading them, by typing:\n')
    print('\t$ bioshed download encode --list\n')
    
    print('You can specify a different output directory, including an AWS S3 remote folder:\n')
    print('\t$ bioshed download encode --output s3://my/output/folder\n')
    
    print('By default, existing files in the output directory will be overwritten. To download only new files:\n')
    print('\t$ bioshed download encode --update\n')

    print('Successful download will also generate an associated annotation file "annotation_encode.txt".\n')
    return

def convert_to_search_string( args ):
    """ Given a category and search terms, outputs an updated url search string

    terms: breast cancer
    category: general
    ---
    urlstring: search string

    Example: 'breast cancer', 'general' => ?type=Experiment&searchTerm=breast+cancer
    Example: 'colon cancer', 'disease' => ...
    NOTE: there may be overlapping terms, which is why original search string is passed in (may be modified in-place).
    """
    terms = args['terms']
    category = args['category']
    pre_search_file = 'files/search_encode_{}.txt'.format(str(category).lower())
    search_string = ''

    try:
        if os.path.exists(pre_search_file):
            df = pd.read_csv(pre_search_file, sep='\t')
            if category != '' and terms == '':
                # special case: search encode --category where we list valid search terms.
                unique_terms = list(set(list(df['ID'])))
                print('Valid search terms for category {}'.format(category))
                for unique_term in unique_terms:
                    if unique_term != '.':
                        print('\t{}'.format(unique_term))
                raise ValueError('Search not performed.')
            elif 'link' in df.columns and 'ID' in df.columns:
                pd_query = df[df['ID']==terms]['link'] # df.query('ID=={}'.format(terms))['link']
                if len(pd_query) > 0:
                    search_string = pd_query.values[0]
        elif not os.path.exists(pre_search_file) and category not in ['general', ''] and terms == '':
            # incorrect category used
            raise ValueError('ERROR: Invalid search category {}. Type "bioshed search encode" to see valid search categories, or just search without categories.'.format(category))

        if search_string == '':
            search_string = '?type=Experiment&searchTerm={}'.format(terms.lower().replace(' ','+'))
        return search_string
    except ValueError as e:
        print(e)
        return ''

def combine_search_strings( ss1, ss2 ):
    """ Combine two search strings into a single ENCODE search string.
    This removes any redundant search terms.
    """
    if 'type=Experiment' in ss1 and 'type=Experiment' in ss2:
        ss2 = ss2.replace('type=Experiment', '')
    if 'type=Biosample' in ss1 and 'type=Biosample' in ss2:
        ss2 = ss2.replace('type=Biosample', '')

    if ss1 != '' and ss2 != '':
        return '{}&{}'.format(ss1,ss2.lstrip('?'))
    elif ss1 == '' and ss2 != '':
        return ss2
    elif ss1 != '' and ss2 == '':
        return ss1
    else:
        return ''

########################## DEPRECATED ############################

def search_encode_general( args ):
    """ Search ENCODE for datasets using general search terms.
    tissue: same as organ...
    celltype: ...
    assaytype: ...
    species: ...
    filetype: ...
    platform: Illumina etc..
    generic: generic search...
    returntype: 'raw' (default), 'full', 'file', 'experiment', 'assay', 'tissue',...
    ---
    results:

    EXAMPLE: search_encode( dict(tissue='breast cancer'))

    https://www.encodeproject.org/search/?type=Experiment&searchTerm=breast+cancer

    >>> search_encode( dict(tissue='breast cancer', returntype='full'))
    ''
    >>> search_encode( dict(tissue='breast cancer', returntype='experiment'))
    ''
    >>> search_encode( dict(tissue='breast cancer', returntype='assay'))
    ''
    """
    returntype = args['returntype'] if 'returntype' in args else 'full'
    search_args = ''
    for arg in args:
        if arg not in ['returntype']:
            search_args += args[arg].replace(' ','%20') if arg in args else ''
    search_url = '/search/?type=Experiment&searchTerm={}'.format(search_args)
    results = encode_search_url( dict(url=search_url, searchtype='full', returntype=returntype))
    print_dataframe( results )
    return results



import os, sys, json
SCRIPT_DIR = str(os.path.dirname(os.path.realpath(__file__)))
HOME_PATH = os.path.expanduser('~')
INIT_PATH = os.path.join(HOME_PATH, '.bioshedinit/')
BIOCONTAINERS_REGISTRY = 'public.ecr.aws/biocontainers'

sys.path.append(os.path.join(SCRIPT_DIR))
import bioshed_core_utils
import bioshed_init
import bioshed_deploy_core
sys.path.append(os.path.join(SCRIPT_DIR, 'bioshed_utils/'))
import docker_utils
import aws_batch_utils
import quick_utils
sys.path.append(os.path.join(SCRIPT_DIR, 'bioshed_atlas/'))
import atlas_encode_utils
import atlas_tcga_utils

AWS_CONFIG_FILE = os.path.join(INIT_PATH,'aws_config_constants.json')
GCP_CONFIG_FILE = ''
PROVIDER_FILE = os.path.join(INIT_PATH, 'hs_providers.tf')
MAIN_FILE = os.path.join(INIT_PATH, 'main.tf')
SYSTEM_TYPE = bioshed_core_utils.detect_os()
VALID_COMMANDS = ['init', 'setup', 'build', 'run', 'runlocal', 'deploy', 'search', 'download', 'teardown', 'keygen']
VALID_PROVIDERS = ['aws', 'amazon', 'gcp', 'google']

def bioshed_cli_entrypoint():
    bioshed_cli_main( sys.argv )
    return

def bioshed_cli_main( args ):
    """ Main function for parsing command line arguments and running stuff.
    args: list of command-line args

    [TODO] figure out local search
    [DONE] add docker installation to "pip install bioshed"
    """
    ogargs = quick_utils.format_type(args, 'space-str')       # original arguments

    if SYSTEM_TYPE == 'unsupported' or SYSTEM_TYPE == 'windows': # until I can support windows
        print('Unsupported system OS. Linux (Ubuntu, Debian, RedHat, AmazonLinux) or Mac OS X currently supported.\n')
    elif len(args) > 1:
        cmd = args[1].strip()
        if cmd in ['run', 'runlocal'] and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) < 3:
                print('You must specify a module with at least one argument - ex: bioshed run fastqc -h')
                print('Type: "bioshed run --help" for full documentation.')
                return

            # special case: CMD --example -> always run locally
            if args[-1] == '--example':
                if cmd == 'run' and '--local' not in args:
                    args = args[0:2] + ['--local'] + args[2:-1]

            # special case: CMD --help -> always run locally
            if args[-1] == '--help' and len(args) > 3 and args[-2] not in ['run', 'runlocal'] and 'biocontainers' not in ogargs:
                if cmd == 'run' and '--local' not in args:
                    args = args[0:2] + ['--local'] + args[2:]

            # special case: bioshed run --list => list possible applications
            if args[-1] == '--list' and len(args) == 3 and 'biocontainers' not in ogargs:
                public_containers = docker_utils.list_containers()
                print('Use "bioshed run" to run one of the applications below:\n')
                for public_container in sorted(public_containers):
                    if public_container not in ['test']:
                        print('\t{}'.format(public_container))
                print('')
                print('Type: "bioshed run --help" for full documentation.\n')
                return

            # special case: if no cloud provider is fully setup, then run locally
            if not (bioshed_init.cloud_configured({}) and bioshed_init.cloud_core_setup( dict(configfile=AWS_CONFIG_FILE))):
                args = args[0:2] + ['--local'] + args[2:]

            args = args[2:] # don't need to parse "bioshed run/runlocal"
            dockerargs = ''
            registry = ''
            ctag = ''
            # optional argument is specified
            while args[0].startswith('--'):
                if args[0]=='--aws-env-file':
                    if len(args) < 2:
                        print('Either did not specify env file or module name - ex: bioshed runlocal --aws-env-file .env fastqc -h')
                    dockerargs += '--env-file {} '.format(args[1])
                    args = args[2:]
                elif args[0]=='--local':
                    cmd = 'runlocal'
                    # when running locally, files in current directory get passed into /input/ dir of container
                    current_dir = str(os.getcwd()).replace(' ','\ ')
                    #if '--inputdir' not in ogargs:
                    #    dockerargs += '-v {}:/input/ '.format(current_dir)
                    if 'biocontainers' not in ogargs:
                        args = docker_utils.specify_output_dir( dict(program_args=args[1:], default_dir=current_dir))
                        # if no cloud bucket is specified, then output to local.
                        if 's3://' not in quick_utils.format_type(args, 'space-str') and 'gcp://' not in quick_utils.format_type(args, 'space-str'):
                            dockerargs += '-v {}:/output/:Z '.format(current_dir)
                        if '--aws-env-file' not in ogargs and 's3://' in quick_utils.format_type(args, 'space-str'):
                            # if S3 bucket is specified and aws-env-file is not specified, then use default aws config file
                            args = ['--aws-env-file', bioshed_init.get_env_file(dict(cloud='aws', initpath=INIT_PATH))] + args
                    else:
                        args = args[1:]
                elif args[0]=='--inputdir':
                    if len(args) < 2:
                        print('You need to specify an input directory.')
                    if args[1] == '.':
                        args[1] = '$(pwd)'
                    if 'biocontainers' not in ogargs:
                        dockerargs += '-v {}:/input/ '.format(args[1])
                        args = docker_utils.specify_output_dir( dict(program_args=args[2:], default_dir=args[1]))
                    else:
                        # special case: biocontainers
                        dockerargs += '-v {}:/data/ '.format(args[1])
                elif args[0]=='--help':
                    bioshed_init.bioshed_run_help()
                    return
            module = args[0].strip().lower()

            # special case: biocontainers
            if module.lower() == 'biocontainers':
                if len(args) < 2 or (len(args) == 3 and '--help' in ogargs):
                    bioshed_init.biocontainers_help()
                    return
                module = str(args[1].split(':')[0]).strip().lower()
                registry = BIOCONTAINERS_REGISTRY
                ctag = str(args[1].split(':')[1]) if len(args[1].split(':')) > 1 else 'latest'
                args = args[2:]
                cmd = 'runlocal'  # biocontainers can only be run locally
                if '-v' not in dockerargs and ':/data' not in dockerargs and '--inputdir' not in ogargs:
                    dockerargs += '-v $(pwd):/data/ '
            # special case: CMD --example
            elif ogargs.endswith('--example'):
                args = ['cat', '/example.txt']
            # special case: e.g., fastqc - need to explicitly specify output directory
            elif module.lower() in ['fastqc']:
                args = [args[0]] + ['-o', '/output/'] + args[1:]
            # run module
            if cmd == 'run':
                jobinfo = aws_batch_utils.submit_job_awsbatch( dict(name=module, program_args=args))
                print('SUBMITTED JOB INFO: '+str(jobinfo))
            elif cmd == 'runlocal':
                print('TOTAL COMMAND: {} | {}'.format(str(dockerargs), str(args)))
                print('NOTE: If you get an AWS credentials error, you may need to specify an AWS ENV file: --aws-env-file <.ENV>')
                docker_utils.run_container_local( dict(name=module, args=args, dockerargs=dockerargs, registry=registry, tag=ctag))

        elif cmd == 'build' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) < 3:
                print('You must specify a module to build: bioshed build <MODULE> <ARGS>')
                return
            module = args[2].strip()
            parsed_args = bioshed_core_utils.parse_build_args( args[3:] )
            print('MODULE: '+str(module))
            if 'install' in parsed_args:
                docker_utils.build_container( dict(name=module, requirements=parsed_args.install, codebase=parsed_args.codebase ))
        elif cmd == 'setup' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) > 2:
                cloud_provider = args[2].lower()
                if cloud_provider in ['aws', 'amazon']:
                    bioshed_init.bioshed_setup(dict( cloud=cloud_provider, initpath=INIT_PATH, configfile=AWS_CONFIG_FILE, providerfile=PROVIDER_FILE, mainfile=MAIN_FILE))
                else:
                    print('Provider {} currently not supported.'.format(cloud_provider))
            else:
                print('Must specify a cloud provider - e.g., bioshed setup aws')
        elif cmd == 'init':
            # first create init directory if doesn't exist
            if not os.path.exists(INIT_PATH):
                os.mkdir(INIT_PATH)
            # create config file if doesn't exist
            if not os.path.exists(AWS_CONFIG_FILE):
                with open(AWS_CONFIG_FILE,'w') as fout:
                    fout.write('{}')
            # ask for login name (email)
            login_success = bioshed_init.bioshed_login()
            if login_success["login"]:
                quick_utils.add_to_json(AWS_CONFIG_FILE, {"login": login_success["user"]})
                which_os = bioshed_init.bioshed_init(dict(system=SYSTEM_TYPE, initpath=INIT_PATH))
                print('BioShed initial install complete. Follow-up options are:')
                print('1) Type "bioshed setup aws" and then "bioshed deploy core" to setup AWS infrastructure for Bioshed.')
                print('2) Type "bioshed build <module> <args>" to build a new bioinformatics application module.')
                print('3) Type "bioshed search encode/ncbi/local/etc..." to search a system or repository for datasets.')
                print('')
        elif cmd == 'deploy' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) < 3:
                print('Must specify a resource to deploy - e.g., bioshed deploy core\n')
                return
            deploy_resource = args[2]
            # for now, assume cloud provider is AWS
            provider = 'aws'
            deploy_option = args[3] if len(args) > 3 else ''
            bioshed_deploy_core.bioshed_deploy_core(dict(cloud_provider=provider, initpath=INIT_PATH, configfile=AWS_CONFIG_FILE, deployoption=deploy_option))
        elif cmd == 'teardown' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            r = input('You are going to tear down your entire bioshed infrastructure. Are you sure? y/n: ') or "N"
            # have another credential-based check - ask for AWS credentials or some password
            if r.upper() == "Y":
                bioshed_init.bioshed_teardown( dict(initpath=INIT_PATH))
        elif cmd == 'search' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) < 3:
                print('Specify a system or repository to search and type your search terms. Examples:')
                print('\tbioshed search encode <SEARCH_TERMS>')
                print('\tbioshed search gdc <SEARCH_TERMS>')
                print('\tbioshed search tcga <SEARCH_TERMS>')
                print('\tbioshed search ncbi <SEARCH_TERMS>')
                return
            if str(args[2]).lower() == 'encode':
                search_terms = str(' '.join(args[3:])).strip()
                print('Searching ENCODE for: {}'.format(search_terms))
                atlas_encode_utils.search_encode( dict(searchterms=search_terms))
            elif str(args[2]).lower() in ['tcga', 'gdc']:
                search_terms = str(' '.join(args[3:])).strip()
                print('Searching Genomic Data Commons for: {}'.format(search_terms))
                atlas_tcga_utils.search_gdc( dict(searchterms=search_terms))
            elif str(args[2]).lower() == 'ncbi':
                print('NCBI search coming soon!')
            elif str(args[2]).lower() == 'local':
                print('Local search coming soon!')
            else:
                print('Currently supported searches: encode, tcga, gdc. Coming soon: nbci, local')
        elif cmd == 'download' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) < 3:
                print('Specify a system or repository to download files from. Examples:')
                print('\tbioshed download encode')
                print('\tbioshed download gdc')
                print('\tbioshed download tcga')
                print('\tbioshed download ncbi')
                return
            if str(args[2]).lower() == 'encode':
                atlas_encode_utils.download_encode( dict(downloadstr=str(' '.join(args[3:])).strip()))
            elif str(args[2]).lower() in ['tcga', 'gdc']:
                atlas_tcga_utils.download_gdc( dict(downloadstr=str(' '.join(args[3:])).strip()))
            elif str(args[2]).lower() == 'ncbi':
                print('NCBI download coming soon!')
            elif str(args[2]).lower() == 'local':
                print('Local download coming soon!')
            else:
                print('Currently supported downloads: encode, nbci, local')
        elif cmd == 'keygen' and bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            if len(args) < 3 or (len(args) >=3 and str(args[2]).lower() not in VALID_PROVIDERS):
                print('Specify a valid cloud provider to generate an API key for.')
                print('\tbioshed keygen aws')
                print('\tbioshed keygen gcp')
                return
            if str(args[2]).lower() in ['aws', 'amazon']:
                key_file = bioshed_init.generate_api_key( dict(cloud='aws', configfile=AWS_CONFIG_FILE))
                print_key = bioshed_init.get_public_key( dict(configfile=AWS_CONFIG_FILE))
                print(print_key)
            elif str(args[2]).lower() in ['gcp', 'google']:
                key_file = bioshed_init.generate_api_key( dict(cloud='gcp', configfile=GCP_CONFIG_FILE))
                print_key = bioshed_init.get_public_key( dict(configfile=AWS_CONFIG_FILE))
                print(print_key)
        elif cmd in VALID_COMMANDS and not bioshed_init.userExists( dict(quick_utils.loadJSON(AWS_CONFIG_FILE)).get("login", "") ):
            print('Not logged on. Please type "bioshed init" and login first.')
        else:
            print_help_menu()
    else:
        print_help_menu()
    return

def print_help_menu():
    print('Specify a valid subcommand. Valid subcommands are:\n')
    print('\t$ bioshed init')
    print('\t$ bioshed setup aws')
    print('\t$ bioshed deploy core')
    print('\t$ bioshed teardown aws')
    print('\t$ bioshed keygen aws')
    print('')
    print('\t$ bioshed run')
    print('\t$ bioshed build')
    print('')
    print('\t$ bioshed search encode')
    print('\t$ bioshed search tcga')
    print('\t$ bioshed search gdc')
    print('')
    print('\t$ bioshed download encode')
    print('\t$ bioshed download tcga')
    print('\t$ bioshed download gdc')
    print('')
    return



import os, sys, subprocess
SCRIPT_DIR = str(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(os.path.join(SCRIPT_DIR, 'bioshed_utils/'))
import quick_utils
import aws_s3_utils

def bioshed_deploy_core( args ):
    """
    cloud_provider: aws, gcp,...
    initpath: path to init files
    configfile: AWS config file
    deployoption: option for deployment - test, dryrun,...
    ---
    """
    cloud = args['cloud_provider']
    initpath = args['initpath']
    configfile = args['configfile']
    deployoption = args['deployoption'] if 'deployoption' in args else ''
    if cloud.lower() in ['aws','amazon']:
        if deployoption.lower() == 'test':
            bioshed_deploy_core_aws_test( dict(initpath=initpath))
        else:
            bioshed_deploy_core_aws_public( dict(initpath=initpath, deployoption=deployoption, configfile=configfile))
    return

def bioshed_deploy_core_aws_test( args ):
    """ Deploys test core infrastructure.

    initpath: path to init files
    deployoption: option for deployment - ex: dryrun
    ---
    """
    cwd = os.getcwd()
    INIT_PATH = args['initpath'] # os.path.join(os.getcwd(), 'hsinit')
    deployoption = args['deployoption'] if 'deployoption' in args else ''
    with open(os.path.join(INIT_PATH,'main.tf'),'w') as f:
        f.write(
        """
        resource "aws_instance" "app_server" {
          ami = "ami-830c94e3"
          instance_type = "t2.micro"
          associate_public_ip_address = false
          tags = {
            Name = "ExampleAppServerInstance"
          }
        }

        """
        )
        f.write(
        """
        resource "aws_eip" "lb" {
          instance = aws_instance.app_server.id
          vpc      = false
        }
        """
        )
        f.write(
        """
        resource "aws_iam_role" "test_role" {
        name = "test_role"
        assume_role_policy = jsonencode({
          Version = "2012-10-17"
          Statement = [
            {
              Action = "sts:AssumeRole"
              Effect = "Allow"
              Sid    = ""
              Principal = {
                Service = "ec2.amazonaws.com"
              }
            },
          ]
        })
        tags = {
          tag-key = "tag-value"
        }
      }
      """
      )
    os.chdir(INIT_PATH)
    subprocess.call('terraform plan -out hs_deploy_core.plan', shell=True)
    subprocess.call('terraform apply hs_deploy_core.plan', shell=True)
    os.chdir(cwd)
    return

def bioshed_deploy_core_aws_public( args ):
    """ Deploys core infrastructure of public resources for running bioshed modules.

    initpath: path to init files
    configfile = args['configfile']
    region: region to deploy files
    deployoption = args['deployoption'] if 'deployoption' in args else ''
    ---

    TODO: add to existing TF files instead of overwriting.
    TODO: check if VPC CIDR block is already taken before assigning
    TODO: use generated key and public AMI - set this up
    """
    cwd = os.getcwd()
    INIT_PATH = args['initpath'] # os.path.join(os.getcwd(), 'hsinit')
    configfile = args['configfile']
    deployoption = args['deployoption'] if 'deployoption' in args else ''
    region = args['region'] if 'region' in args else 'us-west-2'
    aws_id = aws_s3_utils.get_aws_id()
    with open(os.path.join(INIT_PATH,'variables.tf'),'w') as f:
        f.write(
        """
        variable "name_prefix" {
          type        = string
          description = "Naming prefix for resources"
          default     = "bioshed-managed"
        }

        variable "aws_region" {
          type        = string
          description = "Region for AWS Resources"
        """
        )
        f.write('  default = "{}"\n'.format(region))
        f.write('}\n\n')
        f.write(
        """
        variable "aws_azs" {
          type        = list(string)
          description = "AWS Available Zones to use"
        """
        )
        f.write('  default     = ["{}a", "{}b", "{}c", "{}d"]\n'.format(region, region, region, region))
        f.write('}\n\n')
        f.write(
        """
        # assume that CIDR block is not taken
        variable "vpc_cidr_block" {
          type        = string
          description = "Base CIDR Block for VPC"
          default     = "10.35.0.0/16"
        }

        variable "vpc_cidr_range_public_subnets" {
          type        = list(string)
          description = "CIDR ranges for public subnets"
          default     = ["10.35.0.0/17"]
        }

        variable "enable_dns_hostnames" {
          type        = bool
          description = "Enable DNS hostnames in VPC"
          default     = true
        }

        variable "vpc_subnet_count" {
          type        = number
          description = "Number of private or public subnets to create in VPC"
          default     = 1
        }

        """
        )
        f.write(
        """
        variable "public_ecs_batch_service_role_policy_arns" {
          type        = list(string)
          description = "IAM Role Policies for Batch to be able to start instances"
          default     = ["arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryFullAccess", "arn:aws:iam::aws:policy/AmazonElasticContainerRegistryPublicReadOnly", "arn:aws:iam::aws:policy/AmazonS3FullAccess", "arn:aws:iam::aws:policy/CloudWatchFullAccess", "arn:aws:iam::aws:policy/AmazonECS_FullAccess", "arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceRole", "arn:aws:iam::aws:policy/service-role/AWSBatchServiceRole"]
        }

        variable "bioshed_ecs_instance_role_policy_arn" {
          type        = list(string)
          description = "IAM Role Policiy to attach to instances run within Batch"
          default     = ["arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceforEC2Role", "arn:aws:iam::aws:policy/AmazonS3FullAccess", "arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryFullAccess", "arn:aws:iam::aws:policy/AmazonElasticContainerRegistryPublicReadOnly"]
        }

        # use public ID later
        variable "batch_ami_id_tiny" {
          type        = string
          description = "AMI for Batch EC2 instances with 100GB storage"
          default     = "ami-08576a4860f85fa6d"
        }

        # use aws_key_pair.deployer later
        variable "batch_ec2_key" {
          type        = string
          description = "EC2 key pair for Batch instances"
          default     = "npi_aws_batch"
        }

        """
        )
    with open(os.path.join(INIT_PATH,'main.tf'),'w') as f:
        f.write(
        """
        # Setup VPC and subnets and route tables
        resource "aws_vpc" "vpc" {
          cidr_block           = var.vpc_cidr_block
          enable_dns_hostnames = var.enable_dns_hostnames
          tags = { Name = "${var.name_prefix}-vpc" }
        }

        resource "aws_internet_gateway" "igw" {
          vpc_id = aws_vpc.vpc.id
          tags = { Name = "${var.name_prefix}-igw" }
        }

        resource "aws_subnet" "public_subnets" {
          count                   = var.vpc_subnet_count
          cidr_block              = var.vpc_cidr_range_public_subnets[count.index]
          vpc_id                  = aws_vpc.vpc.id
          map_public_ip_on_launch = true
          availability_zone       = var.aws_azs[count.index]
          tags = { Name = "${var.name_prefix}-public-subnet-${count.index}" }
        }

        # public route table - mapping of VPC CIDR block to local is added automatically
        resource "aws_route_table" "public_route_table" {
          vpc_id = aws_vpc.vpc.id
          route {
            cidr_block = "0.0.0.0/0"
            gateway_id = aws_internet_gateway.igw.id
          }
          tags = { Name = "${var.name_prefix}-public-route-table" }
        }

        # set main route table
        resource "aws_main_route_table_association" "rta-main-public" {
          vpc_id         = aws_vpc.vpc.id
          route_table_id = aws_route_table.public_route_table.id
        }

        # route table associations
        resource "aws_route_table_association" "rta-public-subnets" {
          count          = var.vpc_subnet_count
          subnet_id      = aws_subnet.public_subnets[count.index].id
          route_table_id = aws_route_table.public_route_table.id
        }

        # s3 endpoint inside VPC - routes any S3-bound traffic to this endpoint instead of through NAT->IGW
        resource "aws_vpc_endpoint_route_table_association" "rta-s3-public" {
          route_table_id = aws_route_table.public_route_table.id
          vpc_endpoint_id = aws_vpc_endpoint.s3.id
        }

        resource "aws_vpc_endpoint" "s3" {
          vpc_id = aws_vpc.vpc.id
        """
        )
        f.write('service_name = "com.amazonaws.{}.s3"\n'.format(region))
        f.write('}\n\n')
        f.write(
        """
        # reachable-from-vpc allows full access to a resource if within the VPC.
        resource "aws_security_group" "reachable_from_vpc" {
          name   = "${var.name_prefix}-reachable-from-vpc"
          vpc_id = aws_vpc.vpc.id

          # inbound - full access from inside VPC
          ingress {
            from_port   = 0
            to_port     = 0
            protocol    = "-1"
            cidr_blocks = [var.vpc_cidr_block]
          }

          # outbound - internet access
          egress {
            from_port   = 0
            to_port     = 0
            protocol    = "-1"
            cidr_blocks = ["0.0.0.0/0"]
          }
          tags = { Name = "${var.name_prefix}-reachable-from-vpc" }
        }

        # reachable-with-ssh allows external SSH access for a resource in a public subnet.
        resource "aws_security_group" "reachable_with_ssh" {
          name   = "${var.name_prefix}-reachable-with-ssh"
          vpc_id = aws_vpc.vpc.id

          # inbound - full access from inside VPC
          ingress {
            from_port   = 22
            to_port     = 22
            protocol    = "tcp"
            cidr_blocks = ["0.0.0.0/0"]
          }

          # outbound - internet access
          egress {
            from_port   = 0
            to_port     = 0
            protocol    = "-1"
            cidr_blocks = ["0.0.0.0/0"]
          }

          tags = { Name = "${var.name_prefix}-reachable-with-ssh" }
        }

        """
        )
        f.write(
        """
        # Batch role and policy
        resource "aws_iam_role" "bioshed_aws_batch_service_role" {
          name = "bioshed_aws_batch_service_role"
          assume_role_policy = <<EOF
{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Action": "sts:AssumeRole",
      "Effect": "Allow",
      "Principal": {
        "Service": "batch.amazonaws.com"
      }
    }
  ]
}
EOF
        }

        resource "aws_iam_role_policy_attachment" "bioshed_aws_batch_service_role" {
          role       = aws_iam_role.bioshed_aws_batch_service_role.name
          policy_arn = "arn:aws:iam::aws:policy/service-role/AWSBatchServiceRole"
        }

        """
        )
        f.write(
        """
        # gives EC2 instances deployed by Batch permissions to access container regsitry, container agent, and S3.
        resource "aws_iam_role" "bioshed_ecs_instance_role" {
          name = "bioshed_ecs_instance_role"

          assume_role_policy = <<EOF
{
  "Version": "2012-10-17",
    "Statement": [
      {
        "Action": "sts:AssumeRole",
        "Effect": "Allow",
        "Principal": {
          "Service": ["ec2.amazonaws.com"]
        }
      }
    ]
}
EOF
        }

        resource "aws_iam_role_policy_attachment" "bioshed_ecs_instance_role" {
          role       = aws_iam_role.bioshed_ecs_instance_role.name
          count      = length(var.bioshed_ecs_instance_role_policy_arn)
          policy_arn = var.bioshed_ecs_instance_role_policy_arn[count.index]
        }

        resource "aws_iam_instance_profile" "bioshed_ecs_instance_role" {
          name = "bioshed_ecs_instance_role"
          role = aws_iam_role.bioshed_ecs_instance_role.name
        }

        """
        )
        f.write(
        """
        # Deploys the batch environment within the VPC setup
        resource "aws_batch_compute_environment" "batch_compute_public" {
          compute_environment_name = "${var.name_prefix}_batch_compute_public"

          compute_resources {
            # compute instances within Batch environment will have resource access governed by ECS instance role
            instance_role = aws_iam_instance_profile.bioshed_ecs_instance_role.arn
            instance_type = ["m5.large", "m5.xlarge", "m5.2xlarge", "m5.4xlarge"]

            max_vcpus = 256
            min_vcpus = 0

            security_group_ids = [
              aws_security_group.reachable_from_vpc.id,
              aws_security_group.reachable_with_ssh.id
            ]

            # ec2_key_pair = var.batch_ec2_key
            ec2_key_pair = aws_key_pair.deployer.id

            subnets = [for s in aws_subnet.public_subnets : s.id]

            type = "EC2"

            image_id = var.batch_ami_id_tiny
            # ec2_configuration {
            #  image_id_override = var.batch_ami_id_tiny
            #  image_type = "ECS_AL2"
            # }
          }

          service_role = aws_iam_role.bioshed_aws_batch_service_role.arn
          type         = "MANAGED"
          depends_on   = [aws_iam_role_policy_attachment.bioshed_aws_batch_service_role, aws_iam_instance_profile.bioshed_ecs_instance_role, aws_internet_gateway.igw, aws_security_group.reachable_from_vpc, aws_security_group.reachable_with_ssh]
        }

        resource "aws_batch_job_queue" "batch_job_queue_public" {
          name     = "${var.name_prefix}_batch_job_queue_public"
          state    = "ENABLED"
          priority = 1
          compute_environments = [
            aws_batch_compute_environment.batch_compute_public.arn
          ]
          depends_on  = [aws_batch_compute_environment.batch_compute_public]
        }

        """
        )
        f.write(
        """
        # ECS batch service role and policy - allows Batch access to containers and container agent.
        resource "aws_iam_role" "bioshed_ecs_batch_service_role" {
          name = "bioshed_ecs_batch_service_role"

          assume_role_policy = <<EOF
{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Action": "sts:AssumeRole",
      "Effect": "Allow",
      "Principal": {
        "Service": ["ec2.amazonaws.com", "ecs.amazonaws.com", "ecs-tasks.amazonaws.com"]
      }
    }
  ]
}
EOF
        }

        resource "aws_iam_role_policy_attachment" "bioshed_ecs_batch_service_role" {
          role       = aws_iam_role.bioshed_ecs_batch_service_role.name
          count      = length(var.public_ecs_batch_service_role_policy_arns)
          policy_arn = var.public_ecs_batch_service_role_policy_arns[count.index]
        }
        """
        )
    os.chdir(INIT_PATH)
    subprocess.call('terraform plan -out hs_deploy_core.plan', shell=True)
    if 'dryrun' not in str(deployoption).lower():
        subprocess.call('terraform apply hs_deploy_core.plan', shell=True)
    os.chdir(cwd)
    # add roles and other config to aws config file
    quick_utils.add_to_json( configfile, dict(aws_ecr_role='arn:aws:iam::{}:instance-profile/bioshed_ecs_instance_role'.format(aws_id), \
                                              aws_ecs_job_role='arn:aws:iam::{}:role/bioshed_ecs_batch_service_role'.format(aws_id), \
                                              jobqueue='bioshed-managed_batch_job_queue_public', \
                                              working_dir='/home',
                                              core_setup='True'))

    return



import os, sys, json
import pandas as pd
import gzip
SCRIPT_DIR = str(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(os.path.join(SCRIPT_DIR))
import atlas_utils
sys.path.append('bioshed_utils/')
import quick_utils
import aws_s3_utils

GENERIC_TERMS = ["cancer", "tumor", "tumour", "dataset"]
DEFAULT_SEARCH_FILE = "search_gdc.txt"

def search_gdc( args ):
    """ Entrypoint for a TCGA or GDC search.
    In general, we recommend using GDC, as this repository includes TCGA and several other consortiums.

    $ bioshed search tcga <searchterms>
    -OR-
    $ bioshed search gdc <searchterms>

    Examples
    $ bioshed search gdc breast cancer variants
    $ bioshed search gdc --tissue heart --assay rna-seq

    searchterms: search terms input by user
    ---
    search_results: data frame of results

    >>> search_gdc( dict(searchterms='breast cancer variants'))
    ''
    >>> search_gdc( dict(searchterms='--tissue heart --assay rna-seq'))
    ''
    >>> search_gdc( dict(searchterms='--assay single cell --tissue brain'))
    ''

    Prints number of experiment datasets found and where results are output to (search_gdc.txt).
    Outputs search results to search_gdc.txt (tab-delimited text)
    This search results file is fed into download relevant files with the following command:

    $ bioshed download gdc
    -OR-
    $ bioshed download tcga

    You can further filter search_gdc.txt before download by adding search terms:

    $ bioshed download gdc --filetype txt

    By default, this will download to the current folder. You can specify a relative path to download or a remote cloud bucket to download to with the --output parameter.

    $ bioshed download gdc --output <local_outdir>
    $ bioshed download gdc --output s3://my/s3/folder

    NOTE: You MUST have a bioshed_gdc.txt file before you run bioshed download.

    You can also specify a bioshed-formatted GDC results file to download files:

    $ bioshed download gdc newsearch_results.txt

    For help with anything, type:
    $ bioshed search gdc --help
    $ bioshed download gdc --help
    """
    MANIFEST_FILE = os.path.join(SCRIPT_DIR, "files/gdc/manifest-all-gdc.txt.gz")
    CATEGORIES_FILE = os.path.join(SCRIPT_DIR, 'files/gdc/categories-all-gdc.txt')
    print('Using manifest file: {}'.format(MANIFEST_FILE))
    print('Using category file: {}'.format(CATEGORIES_FILE))
    search_results = {}
    search_string = args['searchterms'] if 'searchterms' in args else ''
    # dictionary of search terms: {"general": "...", "tissue": "...", "celltype": "..."...}
    search_dict = atlas_utils.parse_search_terms( search_string ) if search_string != '' else {}
    if search_dict == {} or 'help' in search_dict:
        print_gdc_help()
    else:
        search_dict = convert_general_terms( search_dict, CATEGORIES_FILE )
        print('Search dictionary: {}'.format(str(search_dict)))
        search_results = get_manifest_rows( search_dict, search_string, MANIFEST_FILE )
    return search_results


def convert_general_terms( search_dict, CATEGORIES_FILE ):
    """ Converts general search terms to specific categories.

    search_dict: search dictionary
    CATEGORIES_FILE: tab-delimited file with category in 1st column and possible search terms in 2nd column
    ---
    search_dict: updated search dict

    Example:
    {"general": "breast cancer rna-seq"} => {"tissue": "breast", "assay": "rna-seq"}
    """
    categories = {}
    if "general" in search_dict:
        with open(CATEGORIES_FILE,'r') as f:
            for r in f:
                rt = r.strip().split('\t')
                category = rt[0]
                terms = rt[1]
                categories[category] = terms

        general_terms = search_dict["general"]
        # just use simple word-based search for now
        general_terms_split = quick_utils.format_type(general_terms, 'list')
        for gterm in general_terms_split:
            for k, v in categories.items():
                if quick_utils.quick_format(gterm) in v and gterm not in GENERIC_TERMS:
                    search_dict = atlas_utils.add_term(search_dict, k, gterm)
    return search_dict


def get_manifest_rows( search_dict, search_string, MANIFEST_FILE ):
    """
    Gets rows from GDC-formatted manifest file that match search terms.
    Valid search terms are:
    --tissue / --assay / --celltype / --disease / --filetype / --platform / --species

    search_dict: dictionary of search terms: {"general": "...", "tissue": "...", "assay": "..."}
    search_string: original search string
    ---
    results: data frame of results
    (out): filtered GDC-format manifest file
    """

    df = pd.read_csv(MANIFEST_FILE, compression='gzip', sep='\t')
    columns = list(df.columns)
    # look for each search term within the corresp category column in the manifest data frame
    for k, v in search_dict.items():
        if k == 'celltype' or (k == 'disease' and ('tumor' in v or 'tumour' in v or 'cancer' in v)):
            k = 'tissue'
        if k in columns:
            # special case where user wants a list of valid search terms, e.g. search gdc --filetype
            if len(search_dict) == 1 and v.strip() == "":
                unique_terms = str('_'.join(list(set(list(df[k]))))).split('_')
                print('Valid search terms for category {}'.format(k))
                for unique_term in unique_terms:
                    if unique_term != '.':
                        print('\t{}'.format(unique_term))
                return {}
            # otherwise we do a search
            terms = quick_utils.format_type(v, 'list')
            for t in terms:
                t = quick_utils.quick_format(t)
                if t not in GENERIC_TERMS:
                    df = df.loc[df[k].str.contains(t, case=False)]
    # write filtered data frame to output file for download
    with open(DEFAULT_SEARCH_FILE,'w') as fout:
        fout.write('# bioshed search gdc {}\n'.format(search_string))
    df.index.name = 'index'
    df.to_csv(DEFAULT_SEARCH_FILE, sep='\t', mode='a')

    print(df)
    print('')
    print('Number of files found: {}'.format(str(len(df['filename']))))
    print('Number of unique assays found: {}'.format(str(len(list(set(df['assay']))))))
    print('Number of tissues-of-origin found: {}'.format(str(len(list(set(df['tissue']))))))

    print('Search results written to {}.'.format(DEFAULT_SEARCH_FILE))
    print('Type "bioshed download gdc" to download data files or "bioshed download gdc --list" for file info before downloading.')
    return df

def download_gdc( args ):
    """ Entrypoint for an GDC download.
    Assumes that search_gdc() has already been run, so that a search_gdc.txt file exists.
    User can refine search.

    downloadstr: download string passed in, which can include the following:

    --help (help menu)
    --list : list files, but do not download (like a dryrun)
    --update : update only - do not overwrite existing files
    --input <file name>
    --output <output directory>
    --filetype <refine by file type(s) - space delimited>
    --assay <refine by assay>
    --id <refine by ID>
    --index <refine by index>
    --tissue <refine by tissue of origin>

    Multiple ids or indexes can be specified by either comma or space delimiting:
    --id 34005 34006
    --id 34005,34006

    [TODO] Clean up documentation and write tests

    [NOTE] Current format for search_gdc.txt is:
    id      filename        md5     project assay   tissue  disease species platform        filetype
    6fd3fe64-23ba-4db3-8315-1867bdec277d    b7274dab-7650-4de3-91a1-7accf806e870.mirbase21.isoforms.quantification.txt      a773004e175527ad668a8db92f0c4e80        tcga    transcriptome-mirna-seq-small-rnaseq    heart   .       human   .       filetype-txt
    [NOTE] '|'.join learned from:
    https://stackoverflow.com/questions/26577516/how-to-test-if-a-string-contains-one-of-the-substrings-in-a-list-in-pandas
    """
    BASE_S3_DIR = "s3://tcga-2-open/"
    BASE_HTTPS_DIR = "https://api.gdc.cancer.gov/data/"
    INFO_COLUMNS = ['id', 'assay', 'tissue']
    dd = atlas_utils.parse_search_terms( args['downloadstr'] if 'downloadstr' in args else '')
    infile = dd['input'] if 'input' in dd else DEFAULT_SEARCH_FILE
    outdir = dd['output'] if 'output' in dd else str(os.getcwd())
    filetype = dd['filetype'] if 'filetype' in dd else ''
    assay = dd['assay'] if 'assay' in dd else ''
    _id = dd['id'] if 'id' in dd else ''
    tissue = dd['tissue'] if 'tissue' in dd else ''
    index = dd['index'] if 'index' in dd else ''
    filename = dd['filename'] if 'filename' in dd else ''
    listonly = 'True' if 'list' in dd else ''
    updateonly = 'True' if 'update' in dd else ''
    outfiles = []
    downloaded_files = []
    annotation_info = []
    ANNOTATION_INFO_FILE = os.path.join(os.getcwd(),'annotation_gdc.txt')
    original_search_command = ''

    if 'help' in dd:
        print_gdc_help()
    elif os.path.exists(infile):
        # get comment line (first line), if there
        with open(infile,'r') as f:
            r = f.readline()
            if r[0] == '#':
                original_search_command = r.strip()
        # get table                        
        df = pd.read_csv(infile, sep='\t', comment='#')
        # print(df)
        if assay != '':
            df = df.loc[df['assay'].str.contains(assay, case=False)]
        if _id != '':
            _id = quick_utils.format_type(_id, 'list')
            df = df.loc[df['id'].str.contains('|'.join(_id), case=False)]
        if index != '':
            index = quick_utils.format_type(index, 'list')
            df = df.loc[df['index'].astype(str).str.contains('|'.join(index))]
        if tissue != '':
            df = df.loc[df['tissue'].str.contains(tissue, case=False)]
        if filetype != '':
            df = df.loc[df['filetype'].str.contains(filetype, case=False)]
        if filename != '':
            df = df.loc[df['filename'].str.contains(filename, case=False)]

        # generate list of files
        if quick_utils.cloud_initialized(dict(cloud='aws')):
            df['filepath'] = BASE_S3_DIR + df['id'] + '/' + df['filename']
        else:
            df['filepath'] = BASE_HTTPS_DIR + df['id']
        
        if listonly == 'True':
            # list files, but do not download
            print(df[['filename', 'tissue', 'assay']])
            with open(ANNOTATION_INFO_FILE,'w') as fout:
                if original_search_command != '':
                    fout.write(original_search_command+'\n')
                fout.write('FILE\tEXPERIMENT\tASSAY\tCELLTYPE\tSPECIES\n')
                for idx, row in df.iterrows():
                    fout.write('{}\t{}\t{}\t{}\t{}\n'.format(str(row['filename']), '.', str(row['assay']).replace('_','/'), str(row['tissue']).replace('_','/'), 'human'))

        else:
            outfiles = list(df['filepath'])
            # download files
            if outdir.startswith('s3') and len(outfiles) > 0 and outfiles[0].startswith('s3'):
                # s3 file transfer
                downloaded_files = aws_s3_utils.transfer_file_s3( dict(path=outfiles, outpath=outdir, overwrite='False' if updateonly=='True' else 'True'))
            elif not outdir.startswith('s3') and len(outfiles) > 0 and outfiles[0].startswith('s3'):
                downloaded_files = aws_s3_utils.download_file_s3( dict(path=outfiles, localdir=outdir, overwrite='False' if updateonly=='True' else 'True'))
            elif not outdir.startswith('s3') and len(outfiles) > 0 and outfiles[0].startswith('http'):
                # http download to local
                for idx, row in df.iterrows():
                    outfile = row['filepath']
                    outfile_name = row['filename']
                    print('Downloading {}'.format(outfile_name))
                    # get http(s) file
                    _response = quick_utils.get_request( dict(url=outfile))
                    # write to local file
                    open(os.path.join(outdir, quick_utils.get_file_only(outfile_name)), 'wb').write(_response.content)
                    # save list of downloaded files
                    downloaded_files.append(os.path.join(outdir, quick_utils.get_file_only(outfile_name)))


            # print annotation info
            with open(ANNOTATION_INFO_FILE,'w') as fout:
                if original_search_command != '':
                    fout.write(original_search_command+'\n')
                fout.write('FILE\tEXPERIMENT\tASSAY\tCELLTYPE\tSPECIES\n')
                for idx, row in df.iterrows():
                    fout.write('{}\t{}\t{}\t{}\t{}\n'.format(str(row['filename']), '.', str(row['assay']).replace('_','/'), str(row['tissue']).replace('_','/'), 'human'))

    else:
        print('File {} does not exist. Please first run "bioshed search gdc"'.format(infile))
    return downloaded_files


def combine_all( base_dir ):
    """ Combines all manifest files into one tab-delimited manifest.
    See COLS variable for column names and order.

    base_dir: base directory where GDC-format manifest files reside
    ---
    (out): full manifest file
    (out): category list file

    """
    DIRS = ['assay', 'disease', 'filetype', 'platform', 'tissue', 'project']
    COLS = ['filename', 'md5', 'project', 'assay','tissue','disease','species','platform','filetype']
    _MANIFEST_FILE = 'manifest-all-gdc.txt'
    _CATEGORIES_FILE = 'categories-all-gdc.txt'
    manifest_all = {}
    categories = {}

    for DIR in DIRS:
        files_all = os.listdir(os.path.join(base_dir, DIR))
        manifest_files = list(filter(lambda x: x.endswith('.txt'), files_all))
        for mfile in manifest_files:
            category = DIR
            value = mfile.split('.')[-2]
            # add search term to category list
            if category not in categories:
                categories[category] = []
            categories[category].append(value)
            # add to manifest dictionary
            with open(os.path.join(base_dir, DIR,mfile),'r') as f:
                for r in f:
                    rt = r.strip().split('\t')
                    if rt[0] != 'id':
                        _id = rt[0]
                        _fn = rt[1]
                        _md5 = rt[2]
                        if _id not in manifest_all:
                            manifest_all[_id] = {}
                            manifest_all[_id]['filename'] = _fn
                            manifest_all[_id]['md5'] = _md5
                            manifest_all[_id]['species'] = 'human'
                            manifest_all[_id]['project'] = 'other'
                        if category not in manifest_all[_id]:
                            manifest_all[_id][category] = value
                        else:
                            manifest_all[_id][category] += '_'+value

    # write to manifest file
    with open(_MANIFEST_FILE, 'w') as fout:
        fout.write('id\tfilename\tmd5\tproject\tassay\ttissue\tdisease\tspecies\tplatform\tfiletype\n')
        for mid in list(manifest_all.keys()):
            row = mid+'\t'
            for COL in COLS:
                row += manifest_all[mid][COL]+'\t' if COL in manifest_all[mid] else '.\t'
            fout.write(row.rstrip(' \t')+'\n')
    # write to category file
    with open(_CATEGORIES_FILE,'w') as fout:
        for category in list(categories.keys()):
            fout.write('{}\t{}\n'.format(category, str(categories[category])))

    print(str(categories))
    return _MANIFEST_FILE

def gdc_run_all( manifest_list_file ):
    """ Runs gdc_manifest_full() for all files
    """
    base_dir = quick_utils.get_file_folder(manifest_list_file)
    manifest_full_files = []
    with open(manifest_list_file,'r') as f:
        for r in f:
            rt = r.strip().split('\t')
            manifest_file = rt[0]
            json_file = rt[1]
            fout = gdc_manifest_full( os.path.join(base_dir, json_file), os.path.join(base_dir, manifest_file) )
            manifest_full_files.append(fout)
    print(str(manifest_full_files))
    return

def gdc_manifest_full( gdc_json_file, gdc_manifest_file ):
    """ Given GDC JSON metadata file and GDC manifest file from
    Genomic Data Commons, creates a full manifest file with data format
    and data type info.
    https://portal.gdc.cancer.gov/repository
    Select a primary site, then click on Manifest and click on JSON.
    """
    gdc_json_list = quick_utils.getJSON( gdc_json_file )
    gdc_json_dict = {} # file_name: [data_format, data_category]
    gdc_manifest_out = gdc_manifest_file[:-4]+'.full.txt'
    for gdc_entry in gdc_json_list:
        data_format = gdc_entry["data_format"]
        file_name = gdc_entry["file_name"]
        data_category = gdc_entry["data_category"]
        gdc_json_dict[file_name] = [data_format, data_category]
    with open(gdc_manifest_file,'r') as f, open(gdc_manifest_out,'w') as fout:
        fout.write('id\tfilename\tmd5\tdata_format\tdata_category\n')
        for r in f:
            rt = r.strip().split('\t')
            _id = rt[0]
            _fn = rt[1]
            _md5 = rt[2]
            if _id not in ['id'] and _fn in gdc_json_dict:
                data_format = gdc_json_dict[_fn][0]
                data_category = gdc_json_dict[_fn][1]
                fout.write('\t'.join([_id, _fn, _md5, data_format, data_category])+'\n')
    return gdc_manifest_out


def gdc_json_to_txt( gdc_json_file ):
    """ Converts JSON metadata downloaded from Genomic Data Commons
    to a tab-delimited text.
    https://portal.gdc.cancer.gov/repository
    then click on JSON.

    Format: (example)
        [{
      "data_format": "SVS",
      "cases": [
        {
          "case_id": "a9fe64a9-6d22-4e9f-96f3-f16af7d298f8",
          "project": {
            "project_id": "TCGA-UVM"
          }
        }
      ],
      "access": "open",
      "file_name": "TCGA-V4-A9EA-01Z-00-DX1.DB8360B6-2BF1-4538-AF2C-29EB7186946E.svs",
      "data_category": "Biospecimen",
      "file_size": 1865302441
    },{
      "data_format": "IDAT",
      "cases": [
        {
          "case_id": "15d19ccc-52b8-41f6-b1c1-2cc55691aed5",
          "project": {
            "project_id": "TCGA-UVM"
          }
        }
      ],
      "access": "open",
      "file_name": "b7723059-b441-428c-bb7d-c69ddb22a886_noid_Grn.idat",
      "data_category": "DNA Methylation",
      "file_size": 8095272
    }
    """
    fout_name = gdc_json_file[:-5]+'.csv'
    fout = open(fout_name,'w')
    fout.write('file_name\tdata_category\tdata_format\tcase_ids\tproject_ids\n')
    gdc_json_list = quick_utils.getJSON( gdc_json_file )
    for gdc_entry in gdc_json_list:
        data_format = gdc_entry["data_format"]
        if "cases" in gdc_entry and len(gdc_entry["cases"]) > 0:
            case_ids = ','.join(list(map(lambda x: x["case_id"] if "case_id" in x else '', gdc_entry["cases"])))
            project_ids = ','.join(list(map(lambda x: x["project"]["project_id"] if "project" in x else '', gdc_entry["cases"])))
        else:
            case_ids = ''
            project_ids = ''
        file_name = gdc_entry["file_name"]
        data_category = gdc_entry["data_category"]
        fout.write('\t'.join([file_name, data_category, data_format, case_ids, project_ids])+'\n')
    return fout_name


def print_gdc_help():
    """ Prints the command-line help menu when a user types a wrong input or
    when a user just types "bioshed search gdc".
    """
    print('\nWelcome to GDC/TCGA dataset search and download, powered by BioShed Atlas.\n')

    print('------------------------------------------------------------')
    print('BioShed GDC SEARCH')
    print('------------------------------------------------------------\n')
    print('Usage:')
    print('\tbioshed search gdc <SEARCH>')
    print('\tbioshed search tcga <SEARCH>\n')

    print('Search Examples:')
    print('\tGeneral search:\t\t$ bioshed search tcga breast cancer variant call')
    print('\tSearch by category:\t$ bioshed search gdc --tissue heart --assay rna-seq\n')

    print('Search categories:')
    print('\t--tissue\tTissue where DNA/RNA/protein material was extracted.')
    print('\t--assay\tAssay type performed.') 
    print('\t--disease\tDiagnosed disease type.')
    print('\t--filetype\tType of data file.')
    print('\t--platform\tInstrument platform used.')
    print('\t--species\tHuman, mouse, etc.\n')

    print('To list search terms (example):\n')
    print('\t$ bioshed search gdc --tissue')
    print('')
    print('BioShed will write SEARCH results to a file "search_gdc.txt" in the current directory.\n')

    print('------------------------------------------------------------')
    print('BioShed GDC DOWNLOAD')
    print('------------------------------------------------------------\n')
    print('Usage: To download data files listed in "search_gdc.txt", type\n')
    print('\t$ bioshed download gdc\n')

    print('Further refine the files you want using any of the search categories. For example\n')
    print('\t$ bioshed download gdc --filetype fastq')
    print('\t$ bioshed download gdc --id 6fd3fe64-23ba-4db3-8315-1867bdec277d\n')
    
    print('You can list the files before downloading them, by typing:\n')
    print('\t$ bioshed download gdc --list\n')
    
    print('You can specify a different output directory, including an AWS S3 remote folder:\n')
    print('\t$ bioshed download gdc --output s3://my/output/folder\n')
    
    print('By default, existing files in the output directory will be overwritten. To download only new files:\n')
    print('\t$ bioshed download gdc --update\n')
    print('Successful download will also generate an associated annotation file "annotation_gdc.txt".\n')
    return



import os, boto3, subprocess, uuid, json, datetime, sys
from dateutil.tz import *
SCRIPT_DIR = str(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(SCRIPT_DIR)
import quick_utils

HOME_PATH = os.path.expanduser('~')

if quick_utils.cloud_initialized( dict(cloud='aws')):
    # initialize S3 access
    if os.environ.get('AWS_ACCESS_KEY_ID') not in [None, ''] and os.environ.get('AWS_SECRET_ACCESS_KEY') not in [None, ''] and os.environ.get('HOME') not in [None, '']:
        if not os.path.exists(os.path.join(str(os.environ.get('HOME')), '.aws/')):
            os.mkdir(os.path.join(str(os.environ.get('HOME')), '.aws/'))
        if not os.path.exists(os.path.join(str(os.environ.get('HOME')), '.aws/credentials')):
            with open(os.path.join(str(os.environ.get('HOME')), '.aws/credentials'),'w') as f:
                f.write('[default]\naws_access_key_id = {}\naws_secret_access_key = {}\n'.format(str(os.environ.get('AWS_ACCESS_KEY_ID')), str(os.environ.get('AWS_SECRET_ACCESS_KEY'))))

s3 = boto3.resource('s3')
s3_client = boto3.client('s3')
    
def _findMatches(f, patterns, matchAll = False):
    """ Wrapper for _findMatch to search for multiple patterns. If matchAll is True, must match all patterns.
    f: file name, e.g. 'hello.txt', STR
    patterns: patterns to search, e.g. '^.txt', ['^bam', 'myfile'], STR or LIST
    matchAll: do we need to match all patterns? BOOL
    RETURN: whether file matches patterns - True/False BOOL

    >>> _findMatches('myfile.txt', '^.txt')
    True
    >>> _findMatches('myfile.txt', 'myf^')
    True
    >>> _findMatches('myfile.txt', 'file^')
    False
    >>> _findMatches('myfile.bam', 'bam')
    True
    >>> _findMatches('myfile.txt', ['^.txt', 'super'])
    True
    >>> _findMatches('myfile.txt', ['^.txt', 'super'], True)
    False
    """
    matches = []
    if type(patterns) == str:
        patterns = [patterns] if patterns != '' else []

    # if patterns is empty, we return True by default
    if patterns == []:
        return True

    for p in patterns:
        matches.append(_findMatch(f, p))
    if matchAll == False:
        if True in matches:
            return True
        else:
            return False
    elif matchAll == True:
        if False in matches:
            return False
        else:
            return True


def _findMatch(f, p):
    """ function for searching for a pattern match for a file name f.
    Pattern can have the following:
    'bam': file name f contains 'bam'
    '^.txt': file name f ends in .txt
    'myfile^': file name begins with myfile

    f: file name, e.g. 'hello.txt' STR
    p: pattern to search, e.g. '^.txt' STR
    RETURN: if pattern is in file name BOOL

    >>> _findMatch('hello.fastq','^fastq')
    True
    >>> _findMatch('myfile.txt', '^.txt')
    True
    >>> _findMatch('myfile.txt', 'myf^')
    True
    >>> _findMatch('myfile.txt', 'file^')
    False
    >>> _findMatch('myfile.txt', 'file')
    True
    >>> _findMatch('myfile.txt', '')
    True
    """
    # print('FINDMATCH: file {}, pattern {}'.format(str(f), str(p)))
    _isMatch = False
    f = str(f).lower()
    p = str(p).lower()
    # if empty string
    if p == '[]' or p == "['']" or p == '':
        _isMatch = True
    # suffix - file extension at end of filename
    elif p[0]=='^' and p[-1]!='^':
        if f.endswith(p[1:]):
            _isMatch = True
    # prefix - file extension at beginning of filename
    elif p[-1]=='^' and p[0]!='^':
        if f.startswith(p[0:-1]):
            _isMatch = True
    # search pattern somewhere in file extension, separated from base file name by one of [_,-,.]
    elif p.rfind('^') > p.find('^'):
        i = p.find('^')
        j = p.rfind('^')
        if (f[f.find('_'):].find(p[i+1:j]) >= 0) or (f[f.find('.'):].find(p[i+1:j]) >= 0) or (f[f.find('-'):].find(p[i+1:j]) >= 0):
            _isMatch = True
    # search pattern anywhere in file name
    else:
        if f.find(p) >= 0:
            _isMatch = True
    return _isMatch

def get_aws_id():
    """ Uses boto3 to get the AWS account ID.
    """
    return boto3.client("sts").get_caller_identity()["Account"]


def transfer_file_s3( args ):
    """ Transfers object(s) from one S3 path to another S3 folder.

    path: S3 file path(s), STR or LIST
    outpath: S3 output path. Must be a folder, not a file extension.
    overwrite: 'True'/'False' Overwrite if file exists
    ---
    outfiles: S3 path of all output files, STR or LIST

    >>> transfer_file_s3( dict(path='s3://hubscratch/test.R', outpath='s3://hubscratch2/'))
    Transfering from s3://hubscratch/test.R to s3://hubscratch2/
    copy: s3://hubscratch/test.R to s3://hubscratch2/test.R
    's3://hubscratch2/test.R'

    [TODO] try except for subprocess
    """
    s3path = args['path']
    outpath = args['outpath']
    overwrite = args['overwrite'] if 'overwrite' in args else 'True'
    return_type = type(s3path)

    s3path = quick_utils.format_type(s3path, 'list')
    outfiles = []

    for s3p in s3path:
        if s3p.startswith('s3://'):
            if overwrite == 'True' or (overwrite=='False' and not file_exists_s3(s3p)):
                print('Transfering from {} to {}'.format(s3p, outpath))
                cmd = ['aws','s3','cp',s3p,outpath]
                subprocess.call(cmd)
            outfiles.append(os.path.join(outpath,quick_utils.get_file_only(s3p)))
        else:
            print('{} needs to start with S3'.format(s3p))

    if return_type == type(''):
        outfiles = quick_utils.format_type(outfiles, 'str')
    return outfiles

def download_file_s3( args ):
    """ Downloads object(s) from S3 to a local file.
        Can be used to download multiple objects if list or comma-delim str is provided.
        Returns full file path of downloaded local file.

    path: S3 file path, s3://bioshed/myfile.bam, STR
    localdir: local directory to download to, /local/dir, STR
    overwrite: 'True'/'False' overwrite existing files (default 'True')
    ---
    localpath: (RETURN) full path of downloaded local file, e.g. '/local/dir/myfile.bam'
        - return type is same as input path type
        - if not a valid S3 path, then an empty string is returned

    >>> download_file_s3( dict(path='s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz', localdir='./testout/'))
    Downloading from S3 - s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz to ./testout/
    './testout/test-R1.fastq.gz'

    >>> download_file_s3( dict(path='s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz,s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz', localdir='./testout/'))
    Downloading from S3 - s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz to ./testout/
    Downloading from S3 - s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz to ./testout/
    './testout/test-R1.fastq.gz,./testout/test-R2.fastq.gz'

    >>> download_file_s3( dict(path=['s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz','s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz'], localdir='./testout/'))
    Downloading from S3 - s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz to ./testout/
    Downloading from S3 - s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz to ./testout/
    ['./testout/test-R1.fastq.gz', './testout/test-R2.fastq.gz']
    """
    s3path = args['path']
    localdir = args['localdir']
    overwrite = args['overwrite'] if 'overwrite' in args else 'True'
    return_type = type(s3path)

    s3path = quick_utils.format_type(s3path, 'list')
    localpath = []

    for s3p in s3path:
        if s3p.startswith('s3://'):
            bucket = s3p.split('/')[2]
            key = '/'.join(s3p.split('/')[3:])

            object_filename = key.split('/')[-1]
            local_filename = os.path.join(localdir, object_filename)

            if overwrite == 'True' or (overwrite=='False' and not file_exists_s3(s3p)):
                print('Downloading from S3 - {} to {}'.format(s3p, localdir))
                s3.Object(bucket,key).download_file(local_filename)
            localpath.append(local_filename)
        else:
            print('{} needs to start with S3'.format(s3p))

    if return_type == type(''):
        localpath = quick_utils.format_type(localpath, 'str')
    return localpath

def download_folder_s3( args ):
    """ Downloads a folder (and sub-folders) from S3 to a local directory.
        Can specify multiple folders.
        Can search for a pattern by passing this into the 'pattern' key
        Returns the local directory containing the download files.

    path: S3 folder path(s), e.g., 's3://bioshed/test/'
    localdir: local directory to download to, e.g. '/local/dir/'
    *pattern: restrict download to files that match a certain pattern (e.g., '*.fastq.gz')
    *excludepattern: do not download files that match a certain pattern (e.g., '*.fastq.gz')
        - can pass a list to pattern
    ---
    localdir: (RETURN) full path of local folder, e.g. '/local/dir/'

    [TODO] try except for subprocess

    >>> download_folder_s3( dict(path='s3://hubpublicinternal/test/aws_s3_utils/', localdir='./testout/'))
    ['aws', 's3', 'cp', '--recursive', 's3://hubpublicinternal/test/aws_s3_utils/', './testout/']
    download: s3://hubpublicinternal/test/aws_s3_utils/test.create_fastq.log to testout/test.create_fastq.log
    download: s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz to testout/test-R1.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz to testout/test-R2.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload-R2.fastq.gz to testout/test-upload-R2.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload-R1.fastq.gz to testout/test-upload-R1.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload.create_fastq.log to testout/test-upload.create_fastq.log
    './testout/'

    >>> download_folder_s3( dict(path=['s3://hubpublicinternal/test/aws_s3_utils/', 's3://hubpublicinternal/test/file_utils/run_test1/sample_test1/mpileup/'], localdir='./testout2/'))
    ['aws', 's3', 'cp', '--recursive', 's3://hubpublicinternal/test/aws_s3_utils/', './testout2/']
    download: s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz to testout2/test-R1.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz to testout2/test-R2.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload-R1.fastq.gz to testout2/test-upload-R1.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload.create_fastq.log to testout2/test-upload.create_fastq.log
    download: s3://hubpublicinternal/test/aws_s3_utils/test.create_fastq.log to testout2/test.create_fastq.log
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload-R2.fastq.gz to testout2/test-upload-R2.fastq.gz
    ['aws', 's3', 'cp', '--recursive', 's3://hubpublicinternal/test/file_utils/run_test1/sample_test1/mpileup/', './testout2/']
    download: s3://hubpublicinternal/test/file_utils/run_test1/sample_test1/mpileup/mpileup.sample_test1.job.log to testout2/mpileup.sample_test1.job.log
    download: s3://hubpublicinternal/test/file_utils/run_test1/sample_test1/mpileup/sample_test1.mpileup.pileup to testout2/sample_test1.mpileup.pileup
    ['./testout2/']

    >>> download_folder_s3( dict(path='s3://hubpublicinternal/test/aws_s3_utils/', localdir='./testout2/', pattern='*.fastq.gz'))
    aws s3 cp --recursive s3://hubpublicinternal/test/aws_s3_utils/ ./testout2/ --exclude "*" --include "*.fastq.gz"
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload-R1.fastq.gz to testout2/test-upload-R1.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-R2.fastq.gz to testout2/test-R2.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-R1.fastq.gz to testout2/test-R1.fastq.gz
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload-R2.fastq.gz to testout2/test-upload-R2.fastq.gz
    './testout2/'

    >>> download_folder_s3( dict(path='s3://hubpublicinternal/test/aws_s3_utils/', localdir='./testout2/', excludepattern='*.fastq.gz'))
    aws s3 cp --recursive s3://hubpublicinternal/test/aws_s3_utils/ ./testout2/ --exclude "*.fastq.gz"
    download: s3://hubpublicinternal/test/aws_s3_utils/test-upload.create_fastq.log to testout2/test-upload.create_fastq.log
    download: s3://hubpublicinternal/test/aws_s3_utils/test.create_fastq.log to testout2/test.create_fastq.log
    './testout2/'
    """
    s3path = args['path']
    localdir = args['localdir'].rstrip('/')+'/'
    searchpattern = args['pattern'] if 'pattern' in args else ''
    excludepattern = args['excludepattern'] if 'excludepattern' in args else ''
    return_type = type(s3path)

    s3path = quick_utils.format_type(s3path, 'list')
    searchpattern = quick_utils.format_type(searchpattern, 'list')
    excludepattern = quick_utils.format_type(excludepattern, 'list')

    for s3p in s3path:
        if s3p.startswith('s3://'):
            if searchpattern in ['',[]] and excludepattern in ['',[]]:
                cmd = ['aws','s3','cp','--recursive',s3p.rstrip('/')+'/',localdir]
                print(cmd)
                subprocess.check_call(cmd)
            else:
                cmd = 'aws s3 cp --recursive {} {}'.format(s3p, localdir)
                if searchpattern not in ['',[]]:
                    cmd += ' --exclude "*"'
                    for pattern in searchpattern:
                        pattern_with_quotes = '"'+pattern+'"'
                        cmd += ' --include {}'.format(pattern_with_quotes)
                elif excludepattern not in ['',[]]:
                    for pattern in excludepattern:
                        pattern_with_quotes = '"'+pattern+'"'
                        cmd += ' --exclude {}'.format(pattern_with_quotes)
                print(cmd)
                subprocess.call(cmd, shell=True)
    return localdir

def upload_file_s3( args ):
    """ Securely uploads local file(s) to a path in S3.
        Full path of localfile should be specified in the input.

    localfile: local file(s) to upload
    path: S3 folder path to upload to, e.g., 's3://bioshed/test/'

    >>> upload_file_s3( dict(localfile='./testout/test-R1.fastq.gz', path='s3://hubpublicinternal/test/upload/'))
    Uploading to s3 - ./testout/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/
    's3://hubpublicinternal/test/upload/test-R1.fastq.gz'

    >>> upload_file_s3( dict(localfile='./testout/test-R1.fastq.gz,./testout/test-R2.fastq.gz', path='s3://hubpublicinternal/test/upload/'))
    Uploading to s3 - ./testout/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/
    Uploading to s3 - ./testout/test-R2.fastq.gz to s3://hubpublicinternal/test/upload/
    's3://hubpublicinternal/test/upload/test-R1.fastq.gz,s3://hubpublicinternal/test/upload/test-R2.fastq.gz'

    >>> upload_file_s3( dict(localfile=['./testout/test-R1.fastq.gz', './testout/test-R2.fastq.gz'], path='s3://hubpublicinternal/test/upload/'))
    Uploading to s3 - ./testout/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/
    Uploading to s3 - ./testout/test-R2.fastq.gz to s3://hubpublicinternal/test/upload/
    ['s3://hubpublicinternal/test/upload/test-R1.fastq.gz', 's3://hubpublicinternal/test/upload/test-R2.fastq.gz']
    """
    s3path = args['path']
    localfile = args['localfile']
    return_type = type(localfile)

    localfile = quick_utils.format_type(localfile, 'list')
    remotefile = []

    if s3path.startswith('s3://'):
        for f in localfile:
            print('Uploading to s3 - {} to {}'.format(str(f), str(s3path)))
            bucket = s3path.split('/')[2]
            key = os.path.join('/'.join((s3path.rstrip('/')+'/').split('/')[3:-1]),f.split('/')[-1])
            response = s3.Object(bucket,key).upload_file(f, ExtraArgs=dict(ServerSideEncryption='AES256'))
            # maybe check response here
            remotefile.append(os.path.join(s3path, quick_utils.get_file_only(f)))

    if return_type == type(''):
        remotefile = quick_utils.format_type(remotefile, 'str')

    return remotefile

def upload_folder_s3( args ):
    """ Uploads all files in a folder (and sub-folders) from S3 to a local directory.
        Can specify multiple folders.
        Automatically use server-side encryption.
        Returns path to upload folder.

    localfolder: local folder(s) to upload
    path: S3 folder path to upload to, e.g., 's3://bioshed/test/'
    *pattern: restrict upload to files that match a certain pattern (e.g., '*.fastq.gz')
    *excludepattern: do not upload files that match a certain pattern (e.g., '*.fastq.gz')
        - can pass a list to pattern
    ---
    path: S3 upload folder

    >>> upload_folder_s3( dict(localfolder='./testout/', path='s3://hubpublicinternal/test/upload/'))
    ['aws', 's3', 'cp', '--recursive', '--sse', 'AES256', './testout/', 's3://hubpublicinternal/test/upload/']
    upload: testout/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-R1.fastq.gz
    upload: testout/test-upload-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R1.fastq.gz
    upload: testout/test-upload.create_fastq.log to s3://hubpublicinternal/test/upload/test-upload.create_fastq.log
    upload: testout/test.create_fastq.log to s3://hubpublicinternal/test/upload/test.create_fastq.log
    upload: testout/test-upload-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R2.fastq.gz
    upload: testout/test-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-R2.fastq.gz
    's3://hubpublicinternal/test/upload/'

    >>> upload_folder_s3( dict(localfolder='./testout/', path='s3://hubpublicinternal/test/upload/', pattern='*.fastq.gz'))
    ['aws', 's3', 'cp', '--recursive', '--sse', 'AES256', './testout/', 's3://hubpublicinternal/test/upload/', '--exclude', '*', '--include', '*.fastq.gz']
    upload: testout/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-R1.fastq.gz
    upload: testout/test-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-R2.fastq.gz
    upload: testout/test-upload-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R2.fastq.gz
    upload: testout/test-upload-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R1.fastq.gz
    's3://hubpublicinternal/test/upload/'

    >>> upload_folder_s3( dict(localfolder=['./testout/', './testout2/'], path='s3://hubpublicinternal/test/upload/', pattern='*.fastq.gz'))
    ['aws', 's3', 'cp', '--recursive', '--sse', 'AES256', './testout/', 's3://hubpublicinternal/test/upload/', '--exclude', '*', '--include', '*.fastq.gz']
    upload: testout/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-R1.fastq.gz
    upload: testout/test-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-R2.fastq.gz
    upload: testout/test-upload-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R2.fastq.gz
    upload: testout/test-upload-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R1.fastq.gz
    ['aws', 's3', 'cp', '--recursive', '--sse', 'AES256', './testout2/', 's3://hubpublicinternal/test/upload/', '--exclude', '*', '--include', '*.fastq.gz']
    upload: testout2/test-upload-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R1.fastq.gz
    upload: testout2/test-R1.fastq.gz to s3://hubpublicinternal/test/upload/test-R1.fastq.gz
    upload: testout2/test-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-R2.fastq.gz
    upload: testout2/test-upload-R2.fastq.gz to s3://hubpublicinternal/test/upload/test-upload-R2.fastq.gz
    's3://hubpublicinternal/test/upload/'

    >>> upload_folder_s3( dict(localfolder=['./testout/', './testout2/'], path='s3://hubpublicinternal/test/upload/', excludepattern='*.fastq.gz'))
    ['aws', 's3', 'cp', '--recursive', '--sse', 'AES256', './testout/', 's3://hubpublicinternal/test/upload/', '--exclude', '*.fastq.gz']
    upload: testout/test-upload.create_fastq.log to s3://hubpublicinternal/test/upload/test-upload.create_fastq.log
    upload: testout/test.create_fastq.log to s3://hubpublicinternal/test/upload/test.create_fastq.log
    ['aws', 's3', 'cp', '--recursive', '--sse', 'AES256', './testout2/', 's3://hubpublicinternal/test/upload/', '--exclude', '*.fastq.gz']
    upload: testout2/test.create_fastq.log to s3://hubpublicinternal/test/upload/test.create_fastq.log
    upload: testout2/test-upload.create_fastq.log to s3://hubpublicinternal/test/upload/test-upload.create_fastq.log
    upload: testout2/mpileup.sample_test1.job.log to s3://hubpublicinternal/test/upload/mpileup.sample_test1.job.log
    upload: testout2/sample_test1.mpileup.pileup to s3://hubpublicinternal/test/upload/sample_test1.mpileup.pileup
    's3://hubpublicinternal/test/upload/'
    """
    s3path = args['path']
    localfolder = args['localfolder']
    return_type = type(localfolder)
    includepattern = args['pattern'] if 'pattern' in args else ''
    excludepattern = args['excludepattern'] if 'excludepattern' in args else ''

    localfolder = quick_utils.format_type(localfolder, 'list')
    includepattern = quick_utils.format_type(includepattern, 'list')
    excludepattern = quick_utils.format_type(excludepattern, 'list')

    if s3path.startswith('s3://'):
        for f in localfolder:
            cmd = ['aws','s3','cp','--recursive','--sse','AES256',f.rstrip('/')+'/',s3path.rstrip('/')+'/']
            if includepattern not in ['', []]:
                cmd += ['--exclude','*']
                for p in includepattern:
                    cmd += ['--include', p]
            elif excludepattern not in ['', []]:
                for p in excludepattern:
                    cmd += ['--exclude', p]
            print(cmd)
            response = subprocess.check_call(cmd)

    return s3path

def object_exists_s3( args ):
    """ Checks if an object exists in S3.
    path: S3 path
    ---
    exists: (BOOL)
    """
    return False if get_objects_s3( args )==[] else True

def get_objects_s3( args ):
    """ Get objects with full metadata at a given S3 path.
    path: S3 path
    ---
    response: AWS-formatted response containing full metadata for each object
    """
    AWS_CONFIG = quick_utils.loadJSON(os.path.join(HOME_PATH,'.bioshedinit/','aws_config_constants.json')) \
                 if os.path.exists(os.path.join(HOME_PATH,'.bioshedinit/','aws_config_constants.json')) \
                 else quick_utils.loadJSON(os.path.join(SCRIPT_DIR,'aws_config_constants.json'))

    s3path = args['path']

    bucket = s3path.split('/')[2]
    key = str('/'.join(s3path.split('/')[3:])).rstrip('/') + '/'
    region = AWS_CONFIG['aws_region']
    # s3_endpoint = '{}.s3.{}.amazonaws.com/{}'.format(bucket,region,key)
    response = s3_client.list_objects_v2( Bucket=bucket, Prefix=key, Delimiter='/', MaxKeys=5000, StartAfter=key )
    return response

def list_files_s3( args ):
    """ Lists files under a given S3 path.
    path: S3 path
    pattern: file patterns to include ('*.fastq')
    excludepattern: file patterns to exclude ('*.log')
        - patterns can be lists
    """
    s3path = args['path']
    includepattern = args['pattern'] if 'pattern' in args else ''
    excludepattern = args['excludepattern'] if 'excludepattern' in args else ''

    response = get_objects_s3( args )
    files = []

    if "Contents" in response:
        files = list(map(lambda x: str(x['Key']).split('/')[-1], response['Contents']))
        ## narrow down file list if we are searching for a specific file pattern
        if includepattern not in ['',[]] or excludepattern not in ['',[]]:
            files = filter_objects( dict(objects=files, pattern=includepattern, excludepattern=excludepattern))

    return files

def list_folders_s3( args ):
    """ Lists folders under a given S3 path.
    path: S3 path
    pattern: folder patterns to include ('sample*')
    excludepattern: folder patterns to exclude ('*log')
        - patterns can be lists
    """
    s3path = args['path']
    includepattern = args['pattern'] if 'pattern' in args else ''
    excludepattern = args['excludepattern'] if 'excludepattern' in args else ''

    response = get_objects_s3( args )
    files = []

    if "CommonPrefixes" in response:
        files = list(map(lambda x: str(x['Prefix']).strip('/').split('/')[-1]+'/', response['CommonPrefixes']))
        ## narrow down file list if we are searching for a specific file pattern
        if includepattern not in ['',[]] or excludepattern not in ['',[]]:
            files = filter_objects( dict(objects=files, pattern=includepattern, excludepattern=excludepattern))

    return files

def list_objects_s3( args ):
    """ Lists objects (files and folders) under a given S3 path.
    path: S3 path
    pattern: folder patterns to include ('sample*')
    excludepattern: folder patterns to exclude ('*log')
        - patterns can be lists

    >>> list_objects_s3( dict(path='s3://hubtenants/bioshed/', excludepattern='*user'))
    ['jira.pdf', 'readme.txt', 'test/', 'testuser1/', 'testuser2/']
    >>> list_objects_s3( dict(path='s3://hubtenants/bioshed/', excludepattern='testus*'))
    ['jira.pdf', 'readme.txt', 'guestuser/', 'test/']
    >>> list_objects_s3( dict(path='s3://hubtenants/bioshed/', excludepattern='test'))
    ['jira.pdf', 'readme.txt', 'guestuser/']
    >>> list_objects_s3( dict(path='s3://hubtenants/bioshed/', pattern='test'))
    ['test/', 'testuser1/', 'testuser2/']
    >>> list_objects_s3( dict(path='s3://hubtenants/bioshed/', pattern='testuser*'))
    ['testuser1/', 'testuser2/']
    >>> list_objects_s3( dict(path='s3://hubtenants/bioshed/', pattern='*testuser'))
    []
    """
    return list_files_s3( args ) + list_folders_s3( args )

def filter_objects( args ):
    """ Filter list of objects (files or folders) based on pattern

    objects: list of objects
    pattern: pattern of objects to include (e.g., '*.fastq')
    excludepattern: file patterns to exclude ('*.log')
        - patterns can be lists
    ---
    filtered_objects: list of objects after filtering

    """
    objects = args['objects']
    includepattern = args['pattern'] if 'pattern' in args else ''
    excludepattern = args['excludepattern'] if 'excludepattern' in args else ''

    objects = quick_utils.format_type(objects, 'list')
    includepattern = quick_utils.format_type(includepattern, 'list')
    excludepattern = quick_utils.format_type(excludepattern, 'list')

    filtered_objects = []

    if includepattern not in ['',[]]:
        for o in objects:
            for p in includepattern:
                if (p.startswith('*') and o.strip('/').endswith(p.strip('*'))) or \
                   (p.endswith('*') and o.strip('/').startswith(p.strip('*'))) or \
                   (p in o):
                    filtered_objects.append(o)
    elif excludepattern not in ['',[]]:
        for o in objects:
            for p in excludepattern:
                if p.startswith('*'):
                    if not o.strip('/').endswith(p.strip('*')):
                         filtered_objects.append(o)
                elif p.endswith('*'):
                    if not o.strip('/').startswith(p.strip('*')):
                         filtered_objects.append(o)
                else:
                    if p not in o:
                        filtered_objects.append(o)
    return filtered_objects



def get_json_object( s3paths ):
    """ Gets content of JSON files in S3, specified by S3 paths.
    s3paths: string of comma-delimited S3 paths to JSON files - 'obj1,obj2,...'
    returns: corresponding list of JSON content (ordered)
    """
    json_list = []
    s3paths_list = s3paths.split(',')
    for s3path in s3paths_list:
        bucket = s3path.split('/')[2]
        key = '/'.join(s3path.split('/')[3:])

        obj = s3.Object(bucket, key)
        data = obj.get()['Body'].read().decode('utf-8')
        json_data = json.loads(data)
        json_list.append(json_data)
    return json_list

def edit_json_object( s3path, pairs2search, pairs2update ):
    """ Updates an existing S3 JSON object with new pairs.
        pairs2search: list of key:value pairs to search for in JSON object [{key1:value1}, {key2: value2},...]
        pairs2update: matched list of key:value pairs to update in JSON object

    (before: [{"a": 1, "b": 2}, {"c": 3, "d": 4, "e": 5}, {"g": 6}, {"h": 8, "i": 9}] )
    >>> aws_s3_utils.edit_json_object('s3://bioshed-db/bioshed/test.json', [{"c": 3}, {"b": 2}], [{"e": 10}, {"j": 11}])
    {'ResponseMetadata': {'RequestId': 'NR3EHTBYPPNH677N', 'HostId': 'Jt08TAQT1BCJ2TFxncN/L5p39BImVU+0Aq+g7SSRp3ltRT2d0l6DQu+xiApplYXsFtbL9XmiBZQ=', 'HTTPStatusCode': 200, 'HTTPHeaders': {'x-amz-id-2': 'Jt08TAQT1BCJ2TFxncN/L5p39BImVU+0Aq+g7SSRp3ltRT2d0l6DQu+xiApplYXsFtbL9XmiBZQ=', 'x-amz-request-id': 'NR3EHTBYPPNH677N', 'date': 'Thu, 11 Aug 2022 03:57:52 GMT', 'x-amz-version-id': 'u1w6fo3dG9LSnhlX6pw__dYp.drkQ5hZ', 'x-amz-server-side-encryption': 'AES256', 'etag': '"288cbe6243118f973884cfa2d8ed7aae"', 'server': 'AmazonS3', 'content-length': '0'}, 'RetryAttempts': 1}, 'ETag': '"288cbe6243118f973884cfa2d8ed7aae"', 'ServerSideEncryption': 'AES256', 'VersionId': 'u1w6fo3dG9LSnhlX6pw__dYp.drkQ5hZ'}
    (after: [{"a": 1, "b": 2, "j": 11}, {"c": 3, "d": 4, "e": 10}, {"g": 6}, {"h": 8, "i": 9}])
    """
    bucket = s3path.split('/')[2]
    key = '/'.join(s3path.split('/')[3:])
    # current JSON data in object
    json_data = get_json_object( s3path )[0]
    for i in range(0,len(pairs2search)):
        pair = pairs2search[i]
        k2search = list(pair.keys())[0]
        v2search = list(pair.values())[0]
        pair2update = pairs2update[i]
        k2replace = list(pair2update.keys())[0]
        v2replace = list(pair2update.values())[0]
        # now search for k:v pair in json_data - yes, this is O(n^2)
        for j in range(0,len(json_data)):
            json_entry = json_data[j]
            if k2search in json_entry and json_entry[k2search]==v2search:
                json_data[j][k2replace] = v2replace
                break
    response = s3_client.put_object( Body=json.dumps(json_data).encode(), Bucket=bucket, Key=key )
    return response


def add_to_json_object( s3path, newpairs ):
    """ Adds new key value pairs to an existing S3 JSON object

    (before: test.json = [])
    >>> aws_s3_utils.add_to_json_object( "s3://bioshed-db/bioshed/test.json", [{"a": 1, "b": 2}, {"c": 3, "d": 4, "e": 5}])
    {'ResponseMetadata': {'RequestId': 'M2D5NRA0C4XZA1C4', 'HostId': 'tP75NO30/io8AtwCBSI/hkQLytdAQhF8AzQ/cCaPxwy7E+SnBd9dJ9Qw08WsBrV/KJJN8ZU3ZPo=', 'HTTPStatusCode': 200, 'HTTPHeaders': {'x-amz-id-2': 'tP75NO30/io8AtwCBSI/hkQLytdAQhF8AzQ/cCaPxwy7E+SnBd9dJ9Qw08WsBrV/KJJN8ZU3ZPo=', 'x-amz-request-id': 'M2D5NRA0C4XZA1C4', 'date': 'Wed, 10 Aug 2022 23:55:21 GMT', 'x-amz-version-id': '5t9dAbPxN5mWnerG3s1BkOo_KRXJTriT', 'x-amz-server-side-encryption': 'AES256', 'etag': '"818ed2e3dd0b1c240fe0a3cdce52c7d6"', 'server': 'AmazonS3', 'content-length': '0'}, 'RetryAttempts': 1}, 'ETag': '"818ed2e3dd0b1c240fe0a3cdce52c7d6"', 'ServerSideEncryption': 'AES256', 'VersionId': '5t9dAbPxN5mWnerG3s1BkOo_KRXJTriT'}
    (after: test.json = [{"a": 1, "b": 2}, {"c": 3, "d": 4, "e": 5}])
    """
    bucket = s3path.split('/')[2]
    key = '/'.join(s3path.split('/')[3:])
    json_data = get_json_object( s3path )[0]
    # if json_data is itself a list, we can just append. Otherwise we add key by key.
    if type(json_data) == type([]):
        for newpair in newpairs:
            json_data.append( newpair )
    elif type(json_data) == type({}):
        for k, v in newpairs.items():
            if k not in json_data:
                json_data[k] = v

    response = s3_client.put_object( Body=json.dumps(json_data).encode(), Bucket=bucket, Key=key )
    return response



def list_objects_nested( s3path, searchpattern ):
    """ Similar to list_objects but will list objects in nested subfolders
    Nested subfolders are found in the "CommonPrefixes" key:
    {
    "CommonPrefixes": [
        {
            "Prefix": "bioshed/guestuser/"
        },
        {
            "Prefix": "bioshed/test/"
        },...
        ]
    }
    """

    return

def get_metadata( s3paths ):
    """ Gets metadata for objects (files or folders) at the given S3 paths (string-list)
    s3paths: string of comma-delimited S3 paths - 'obj1,obj2,...'
    returns: corresponding list of tags
    """
    metadata_list = []
    s3paths_list = s3paths.split(',')
    for s3path in s3paths_list:
        bucket = s3path.split('/')[2]
        key = '/'.join(s3path.split('/')[3:])
        response = s3_client.get_object_tagging(Bucket=bucket,Key=key)
        if 'TagSet' in response:
            metadata = response['TagSet']
        else:
            metadata = {}
        metadata_list.append( metadata )
    return metadata_list

def set_metadata( s3paths, tags_dict, overwrite = 'True' ):
    """ Sets metadata for objects (files or folders) at the given S3 paths as the passed in tags.
        If tags exist, it overwrites if overwrite = True (default).
        Currently, can have up to 10 tags.
        All objects will get the same tags.

        s3paths: string of comma-delimited objects in S3 - 'obj1,obj2,...'
        tags_dict: DICT - e.g., {'project': 'epigenome', 'id': 'ID-01', 'tissue': 'heart'}
        returns: set tags (list)
    """
    def getCurrentTag( k, tags ):
        """ List of tags format as [{'Key': k, 'Value': v}, {'Key2': k2, 'Value2': v2},...]
            return value for key k
        """
        for t in tags:
            if k == t['Key']:
                return t
        return {}

    tag_sets = []
    s3paths_list = s3paths.split(',')
    for s3path in s3paths_list:
        bucket = s3path.split('/')[2]
        key = '/'.join(s3path.split('/')[3:])
        # get the current tags, and only overwrite if overwrite = True
        current_tags = get_metadata( s3path )[0]
        tag_set = []
        for k, v in tags_dict.items():
            new_tag = {'Key': k, 'Value': v}
            # only overwrite if we specify so, or if tag key does not exist
            if overwrite[0].upper()=='T' or (k not in list(map(lambda kv: kv['Key'], current_tags))):
                tag_set.append(new_tag)
            else:
                # otherwise keep the current tag
                current_tag = getCurrentTag( k, current_tags )
                tag_set.append( current_tag )
        # update metadata for S3 object
        response = s3_client.put_object_tagging(Bucket=bucket,Key=key, Tagging={'TagSet': tag_set})
        tag_sets.append(tag_set)
    return tag_sets

def dateConverter(json_object):
    """ AWS response objects often contain datetime objects.
    For proper JSON stringify, need to convert these datetime objects to string date format.
    """
    if isinstance(json_object, datetime.datetime):
        return json_object.__str__()

####### LEGACY FUNCTIONS (for compatibility) #######
def downloadFile_S3(s3path, dir_to_download):
    return download_file_s3( dict(path=s3path, localdir=dir_to_download))

def downloadFiles_S3(s3paths, dir_to_download):
    return download_file_s3( dict(path=s3paths, localdir=dir_to_download))

def downloadFolder_S3(s3path, localdir):
    return download_folder_s3( dict(path=s3path, localdir=localdir))

def downloadFiles_Pattern_S3(s3_path, directory_to_download, pattern):
    return download_folder_s3( dict(path=s3_path, localdir=directory_to_download, pattern=pattern))

def uploadFile_S3(localfile, s3path):
    return upload_file( dict(path=s3path, localfile=localfile))

def uploadFiles_S3(localfiles, s3path):
    return upload_file( dict(path=s3path, localfile=localfiles))

def uploadFolder_S3(localdir, s3path, files2exclude = ''):
    return upload_folder_s3( dict(path=s3path, localdir=localdir, pattern=files2exclude))

def listSubFiles(s3_path, patterns2include, patterns2exclude):
    """
    Lists files from S3 that match a specific pattern
    :param s3_path: s3 folder path
    :param patterns2include: LIST of file patterns to search for.
    :param patterns2exclude: LIST of file patterns to exclude.
    :return: found files matching pattern

    patterns follow this notation: e.g., ['^.bam', 'hepg2^', 'I1'] where
                 '^.bam' => file ends with BAM
                 'hepg2^' => file begins with hepg2
                 '^fastq^' => file contains fastq in file extension (sep from base file name by [-,_,.]: e.g., myfile.fastq.gz
                 'I1' => file contains the word I1 anywhere

    >>> listSubFiles('s3://hubpublicinternal/test/aws_s3_utils/', 'test', 'R1')
    ['test-R2.fastq.gz', 'test-upload-R2.fastq.gz', 'test-upload.create_fastq.log', 'test.create_fastq.log']
    >>> listSubFiles('s3://hubpublicinternal/test/aws_s3_utils/', 'test', ['^R1^','^R2^'])
    ['test-upload.create_fastq.log', 'test.create_fastq.log']
    """
    if type(patterns2include) == str:
        patterns2include = [patterns2include]
    if type(patterns2exclude) == str:
        patterns2exclude = [patterns2exclude]

    if type(s3_path) == type([]) and s3_path != []:
        s3_path = s3_path[0]
    elif type(s3_path) == type([]) and s3_path == []:
        s3_path = ''

    cmd = 'aws s3 ls %s' % (s3_path.rstrip('/')+'/')
    dfiles = []
    uid = str(uuid.uuid4())[0:6]  # prevents race conditions on tmp file
    # output of S3 copy to temporary file
    try:
        # I think calledprocesserror will be called if file not found
        with open(uid+'_dfilestmptmp.tmp','w') as fout:
            subprocess.check_call(cmd.split(' '), stdout=fout)
        # get a list of all downloaded files
        with open(uid+'_dfilestmptmp.tmp','r') as f:
            for r in f:
                rp = r.split(' ')[-1].lstrip(' \t').rstrip(' \t\n')
                # '.' indicates its a file
                if '.' in rp and _findMatches(rp, patterns2include) and not (patterns2exclude != [] and _findMatches(rp, patterns2exclude)):
                    dfiles.append(rp)

        # remove temporary file
        rm_command = ['rm',uid+'_dfilestmptmp.tmp']
        subprocess.check_call(rm_command)
    except subprocess.CalledProcessError:
        print('CALLED PROCESS ERROR in aws_s3_utils.listSubFiles() or FILES NOT FOUND')
        if os.path.exists(uid+'_dfilestmptmp.tmp'):
            rm_command = ['rm',uid+'_dfilestmptmp.tmp']
            subprocess.check_call(rm_command)
        return []
    return dfiles

def _filter_list_objects_response( response, search_pattern ):
    """ private function
    Narrow down file list response object if we are searching for a specific file pattern
    """
    files2remove = []
    for file_info in response["Contents"]:
        if "Key" in file_info and search_pattern not in file_info["Key"]:
            files2remove.append(file_info)
    for file2remove in files2remove:
        response["Contents"].remove(file2remove)
    return response


def list_objects( s3path, searchpattern = '' ):
    """ List objects in an S3 object path. Returns a JSON response with format:
      {'ResponseMetadata':
         {'RequestId': ...,
          'HostId': ...
          'HTTPStatusCode': 200,
          'HTTPHeaders': ...
          'IsTruncated': False,
          'Contents':
            [{'Key': '<filename>' 'LastModified': ...', 'Size': '...'...},
             {'Key': '...
            ]
         }}}

    searchpattern: only include files that contain the passed searched pattern (optional)

    example of returned file info:
    {"Contents": [
        {
            "Key": "bioshed/jira.pdf",
            "LastModified": "2022-07-27 23:11:22+00:00",
            "ETag": "\"46213a6178ac104ba811cc5bb0133218\"",
            "Size": 159226,
            "StorageClass": "STANDARD"
        },...
    ]}
    """
    AWS_CONFIG = quick_utils.loadJSON(os.path.join(HOME_PATH,'.bioshedinit/','aws_config_constants.json')) \
                 if os.path.exists(os.path.join(HOME_PATH,'.bioshedinit/','aws_config_constants.json')) \
                 else quick_utils.loadJSON(os.path.join(SCRIPT_DIR,'aws_config_constants.json'))

    bucket = s3path.split('/')[2]
    region = AWS_CONFIG['aws_region']
    # s3_endpoint = '{}.s3.{}.amazonaws.com/{}'.format(bucket,region,key)
    if '.' in s3path.split('/')[-1] or len(s3path.rstrip('/').split('/')) < 4:
        # searching for file
        key = str('/'.join(s3path.split('/')[3:])).rstrip('/')
        response = s3_client.list_objects_v2( Bucket=bucket, Prefix=key, Delimiter='/')
    else:
        # searching within a folder
        key = str('/'.join(s3path.split('/')[3:])).rstrip('/') + '/'
        response = s3_client.list_objects_v2( Bucket=bucket, Prefix=key, Delimiter='/', StartAfter=key )
    ## narrow down file list if we are searching for a specific file pattern
    if searchpattern != '' and "Contents" in response:
        return _filter_list_objects_response( response, searchpattern )
    else:
        return response

def listSubFolders(s3_path, folders2include = [], folders2exclude = [], options = ''):
    """ Lists all immediate subfolders under a given S3 path.
    :param s3_path: s3 folder path
    :param folders2include: LIST, if specified, only include these folders
    :param folders2exclude: LIST of folders to exclude
    :param options: options to include with ls call
    :return: found subfolders
    >>> listSubFolders('s3://hubpublicinternal/test/', ['aws_s3_utils'])
    ['aws_s3_utils']
    """
    if type(s3_path) == type([]) and s3_path != []:
        s3_path = s3_path[0]
    elif type(s3_path) == type([]) and s3_path == []:
        s3_path = ''
    cmd = 'aws s3 ls {} {}'.format(options, (s3_path.rstrip('/')+'/'))
    dfolders = []
    uid = str(uuid.uuid4())[0:6]  # prevents race conditions on tmp file
    try:
        fout = open(uid+'_dfolderstmptmp.tmp','w')
        subprocess.check_call(cmd.split(' '), stdout=fout)
        fout.close()
        with open(uid+'_dfolderstmptmp.tmp','r') as f:
            for r in f:
                rp = r.lstrip(' \t').rstrip(' \t\n')
                if rp.startswith('PRE'):  # tags a folder - hopefully this doesn't change on AWS side with aws s3 ls
                    folder = rp.split(' ')[1].rstrip('/')
                    if (folder not in folders2exclude) and (folders2include == [] or folder in folders2include):
                        dfolders.append(folder)

        # remove temporary file
        rm_command = ['rm',uid+'_dfolderstmptmp.tmp']
        subprocess.check_call(rm_command)
        return dfolders
    except subprocess.CalledProcessError:
        return []

def file_exists_s3( f ):
    """ Check if a file exists in S3

    >>> file_exists_s3( "s3://www.bioshed.io/index.html" )
    True
    >>> file_exists_s3( "s3://www.bioshed.io/index.blah" )
    False
    """
    objects = list_objects( f )
    if "Contents" in objects and len(objects["Contents"]) > 0:
        return True
    else:
        return False
